from __future__ import annotations

import os
import struct
import warnings
from collections.abc import AsyncGenerator, AsyncIterator, Awaitable, Callable
from contextlib import asynccontextmanager
from functools import partial
from ssl import SSLContext
from types import TracebackType
from typing import Any, Literal, Self

import anyio
import attrs
import structlog
from anyio.abc import ByteStream, TaskGroup
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream

from pg_purepy.connection import (
    AsyncPostgresConnection,
    RollbackTimeoutError,
    _open_connection,
    _open_socket,
)
from pg_purepy.conversion import ArrayConverter, Converter
from pg_purepy.exc import ConnectionForciblyKilledError, ConnectionInTransactionWarning
from pg_purepy.messages import (
    DataRow,
    RecoverableDatabaseError,
)

logger: structlog.stdlib.BoundLogger = structlog.get_logger(name=__name__)


@attrs.define()
class OpenedConnection:
    sock: ByteStream = attrs.field()
    conn: AsyncPostgresConnection = attrs.field()


class PooledDatabaseInterface:
    """
    Connection pool based PostgreSQL interface.

    This is the API you should prefer for any real application as the single connection class
    *cannot* be used for concurrent access to the database.
    """

    def __init__(
        self,
        count: int,
        nursery: TaskGroup,
        conn_fn: Callable[[], Awaitable[tuple[ByteStream, AsyncPostgresConnection]]],
    ):
        self._create_connection = conn_fn
        self._connection_count = count

        self._write: MemoryObjectSendStream[OpenedConnection]
        self._read: MemoryObjectReceiveStream[OpenedConnection]

        self._write, self._read = anyio.create_memory_object_stream[OpenedConnection](
            max_buffer_size=count
        )

        # list of converters used when opening a new connection
        self._converters: set[Converter[Any]] = set()

        # used to add new converters. non-queue.
        self._raw_connections: set[AsyncPostgresConnection] = set()

        self._nursery = nursery

    async def _start(self, count: int) -> None:
        logger.debug("Opening pooled connections", count=count)

        for _ in range(count):
            await self._open_new_connection()

    @property
    def max_connections(self) -> int:
        """
        The maximum number of connections this pool may have idle.
        """

        return self._connection_count

    @property
    def idle_connections(self) -> int:
        """
        The number of the connections that haven't currently been checked out.
        """

        return self._read.statistics().current_buffer_used

    @property
    def waiting_tasks(self) -> int:
        """
        The number of tasks that are currently waiting for a connection to be used.
        """

        return self._read.statistics().tasks_waiting_receive

    @staticmethod
    async def _cancel_query(conn: AsyncPostgresConnection) -> None:
        """
        Cancels the query running on this connection.
        """

        logger.debug("Cancelling query", pid=conn._pid)

        sock = await _open_socket(conn._addr, port=conn._port, ssl_context=conn._ssl_context)

        async with sock:
            # yay magic numbers
            data = struct.pack(">IIII", 16, 80877102, conn._pid, conn._secret_key)
            await sock.send(data)

    async def _open_new_connection(self) -> None:
        sock, conn = await self._create_connection()
        self._raw_connections.add(conn)
        for converter in self._converters:
            conn.add_converter(converter=converter)

        wrapped = OpenedConnection(sock, conn)

        try:
            self._write.send_nowait(wrapped)
        except anyio.WouldBlock:
            self._raw_connections.remove(conn)
            raise RuntimeError(
                "Asked to create a new connection, but there's already too many "
                "connections in the queue!"
            ) from None

    async def __aenter__(self) -> Self:
        return self

    async def _cleanup(self) -> Literal[False]:
        # Cancellation-resistant exit algorithm. This takes inspiration from
        # https://github.com/richardsheridan/trio-parallel/blob/37d0451f632e40fc8fa0bc1646180a83274219a3/trio_parallel/_proc.py#L56-L85
        # using a double shielded nursery setup that will try and gracefully exit the connection,
        # then forcibly disconnect if required.

        gathered: list[OpenedConnection] = []
        while True:
            try:
                gathered.append(self._read.receive_nowait())
            except anyio.WouldBlock:
                break

        assert len(gathered) <= self._connection_count

        swallowed: list[Exception] = []
        forcibly_killed: list[OpenedConnection] = []

        async def kill(o: OpenedConnection) -> None:
            if o.conn.dead:
                try:
                    # This *always* marks the connection as terminated, so even if this fails then
                    # .dead will be True for the second nursery.
                    await o.conn._terminate()
                except Exception as e:
                    swallowed.append(e)
                else:
                    # connection successfully exited, we can continue to trying to close the socket
                    # directly.
                    gathered.remove(o)

            # even if we're subsequently cancelled, this will still always close the socket.
            await o.sock.aclose()

        # step 1: close all connections gracefully. this will try and send the Terminate message,
        # then close the socket, sending the TLS close message as appropriate.
        # This is shielded + a deadline so that external cancellations are ignored but it won't
        # deadlock trying to send to a server that is ignoring us.
        async with anyio.create_task_group() as n1:
            n1.cancel_scope.shield = True
            n1.cancel_scope.deadline = anyio.current_time() + 5

            for conn in gathered:
                n1.start_soon(kill, conn)

        # Step 2: if the deadline expired, we then have to just go "fuck it" and forcibly kill
        # everything. The termination message is never sent because all of the connections are
        # considered dead by here.
        if n1.cancel_scope.cancel_called:
            async with anyio.create_task_group() as n2:
                n2.cancel_scope.shield = True

                for conn in gathered:
                    forcibly_killed.append(conn)
                    n2.start_soon(kill, conn)

        # kill any waiting tasks... this should never be a problem but you honestly never know.
        self._read.close()
        self._write.close()

        if not (swallowed or forcibly_killed):
            return False  # don't suppress error inside context manager

        exceptions: list[Exception] = [*swallowed]
        if forcibly_killed:
            exceptions += [ConnectionForciblyKilledError(i.conn) for i in forcibly_killed]

        raise ExceptionGroup("Error when closing down connection pool!", exceptions)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        return await self._cleanup()

    @asynccontextmanager
    async def checkout_connection(
        self, start_new_transaction: bool = False
    ) -> AsyncIterator[AsyncPostgresConnection]:
        """
        Checks out a single connection from the connection pool.

        This is an *asynchronous context manager* that will automatically clean up the connection
        once the context manager block exits.

        If ``start_new_transaction`` is True, this will issue a new transaction before yielding
        the connection.
        """

        checkout = await self._read.receive()
        logger.debug(
            "Connection acquired", connections_available=self._read.statistics().current_buffer_used
        )

        if checkout.conn.in_transaction:
            warnings.warn(
                f"Connection {checkout.conn} still has a transaction open. Forcing rollback.",
                stacklevel=2,
            )
            await checkout.conn._safely_rollback(None)

        try:
            if start_new_transaction:
                await checkout.conn.execute("BEGIN;")

            yield checkout.conn

            if start_new_transaction:
                await checkout.conn.execute("COMMIT;")

            if checkout.conn.in_transaction:
                warn = ConnectionInTransactionWarning(
                    f"Connection {checkout.conn} is being checked back in with a transaction open. "
                    "Forcing rollback."
                )
                warnings.warn(warn, stacklevel=3)
                await checkout.conn._safely_rollback(None)

        except anyio.get_cancelled_exc_class():
            # open a new conn to psql and cancel the query
            with anyio.move_on_after(delay=5.0, shield=True) as scope:
                await self._cancel_query(conn=checkout.conn)

                try:
                    await checkout.conn.wait_until_ready()
                except RecoverableDatabaseError as e:
                    if e.response.code != "57014":
                        raise

            if scope.cancel_called:
                logger.warning("Failed to cancel query", pid=checkout.conn._pid)

            raise

        finally:
            rollback_failed = None
            if checkout.conn.in_transaction:
                try:
                    await checkout.conn._safely_rollback(None)
                except RollbackTimeoutError as e:
                    rollback_failed = e

            if checkout.conn.dead:
                logger.warning("Pool connection dead", pid=checkout.conn._pid)

                with anyio.CancelScope(shield=True):
                    try:
                        self._raw_connections.remove(checkout.conn)
                        await checkout.sock.aclose()
                    finally:
                        self._nursery.start_soon(self._open_new_connection)
            else:
                try:
                    self._write.send_nowait(checkout)
                except anyio.WouldBlock:
                    raise RuntimeError(
                        "Attempted to send connection back to queue, but there's "
                        "already too many connections in the queue!"
                    ) from None

            if rollback_failed:
                raise rollback_failed

    ## High-level methods. ##
    @asynccontextmanager
    async def checkout_in_transaction(self) -> AsyncGenerator[AsyncPostgresConnection]:
        """
        Shortcut for ``checkout_connection(start_new_transaction=True)``.
        """

        async with self.checkout_connection(start_new_transaction=True) as conn:
            yield conn

    async def execute(self, query: str, *params: Any, **kwargs: Any) -> int:
        """
        Executes a query on the next available connection. See
        :meth:`.AsyncPostgresConnection.execute` for more information.
        """

        async with self.checkout_connection() as conn:
            return await conn.execute(query, *params, **kwargs)

    async def fetch(self, query: str, *params: Any, **kwargs: Any) -> list[DataRow]:
        """
        Fetches the result of a query on the next available connection. See
        :meth:`.AsyncPostgresConnection.fetch` for more information.
        """

        async with self.checkout_connection() as conn:
            return await conn.fetch(query, *params, **kwargs)

    async def fetch_one(
        self,
        query: str,
        *params: Any,
        **kwargs: Any,
    ) -> DataRow:
        """
        Like :meth:`.fetch`, but only returns one row. See
        :meth:`.AsyncPostgresConnection.fetch_one` for more information.
        """

        async with self.checkout_connection() as conn:
            return await conn.fetch_one(query, *params, **kwargs)

    async def fetch_one_or_none(self, query: str, *params: Any, **kwargs: Any) -> DataRow | None:
        """
        See :meth:`.AsyncPostgresConnection.fetch_one_or_none`.
        """

        async with self.checkout_connection() as conn:
            return await conn.fetch_one_or_none(query, *params, **kwargs)

    ## Utility Methods ##
    async def find_oid_for_type(self, type_name: str) -> int | None:
        """
        Finds the OID for the type with the specified name.
        """

        row = await self.fetch_one("select oid from pg_type where typname = :name", name=type_name)
        return row.data[0]

    def add_converter(self, converter: Converter[Any]) -> None:
        """
        Registers a converter for all the connections on this pool.
        """
        self._converters.add(converter)

        for conn in self._raw_connections:
            conn.add_converter(converter)

    async def add_converter_using(
        self,
        fn: Callable[[AsyncPostgresConnection], Awaitable[Converter[Any] | None]],
        with_array: bool = False,
    ) -> None:
        """
        Adds a converter using the specified async function.

        This is useful for dealing with PostgreSQL extensions that register custom types, where
        the OID of such types is not fixed and needs to be collected at runtime.
        """

        async with self.checkout_connection() as conn:
            converter = await fn(conn)

        if converter is not None:
            if with_array:
                await self.add_converter_with_array(converter)
            else:
                self.add_converter(converter)

    async def add_converter_with_array(self, converter: Converter[Any], **kwargs: Any) -> None:
        """
        Registers a converter and also registers the converter for arrays of its type.
        """

        self.add_converter(converter)

        async with self.checkout_connection() as conn:
            row = await conn.fetch_one(
                "select typarray::oid from pg_type where oid = :oid", oid=converter.oid
            )

            arrcv = ArrayConverter(oid=row[0], subconverter=converter, **kwargs)  # type: ignore
            self.add_converter(arrcv)


def determine_conn_count() -> int:
    """
    Determines the appropriate default connection count.
    """

    if count := os.cpu_count():
        return (count * 2) + 1

    return 2  # fuck it.


@asynccontextmanager
async def open_pool(
    address_or_path: str | os.PathLike[str],
    username: str,
    *,
    connection_count: int | None = None,
    port: int = 5432,
    password: str | None = None,
    database: str | None = None,
    ssl_context: SSLContext | None = None,
) -> AsyncGenerator[PooledDatabaseInterface]:
    """
    Opens a new connection pool to a PostgreSQL server. This is an asynchronous context manager.

    This takes the same arguments and keyworrd arguments as
    :func:`.open_database_connection`, except for the optional
    ``connection_count`` parameter.

    :param connection_count: The ideal number of connections to keep open at any one time. The
                             pool may shrink slightly as connections are closed due to network
                             errors and aren't immediately re-opened.

    By default, the connection count is (CPU_COUNT * 2) + 1.
    """

    if connection_count is None:
        connection_count = determine_conn_count()

    conn_fn = partial(
        _open_connection,
        address_or_path=address_or_path,
        username=username,
        port=port,
        password=password,
        database=database,
        ssl_context=ssl_context,
    )

    async with (
        anyio.create_task_group() as tg,
        PooledDatabaseInterface(
            connection_count,
            tg,
            conn_fn=conn_fn,
        ) as pool,
    ):
        await pool._start(connection_count)

        yield pool



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, cast, override

from pg_purepy.conversion.abc import Converter

if TYPE_CHECKING:
    from pg_purepy.protocol import ConversionContext


class EnumConverter[T: Enum](Converter[T]):
    """
    A converter that lets you use Python enums for PostgreSQL enums.
    """

    def __init__(
        self,
        oid: int,
        enum_klass: type[T],
        *,
        use_member_values: bool = False,
        lowercase_names: bool = True,
    ):
        """
        :param oid: The PostgreSQL object ID of the enum type.
        :param enum_klass: The Python-side enumeration class.
        :param use_member_values: If True and the enum values are all strings, then the member
                                  values will be used rather than the names.
        :param lowercase_names: If ``use_member_values`` is False, and this is True, then the
                                member names will be lowercased. This is the general PostgreSQL
                                enum convention.
        """
        self.oid = oid
        self._member_klass = enum_klass
        self._use_members = use_member_values
        self._lowercase_names = lowercase_names

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> T:
        if self._use_members:
            return self._member_klass(data)

        try:
            return self._member_klass[data]
        except KeyError:
            if not self._lowercase_names:
                raise

            # hmm...
            return self._member_klass[data.upper()]

    @override
    def to_postgres(self, context: ConversionContext, data: Enum) -> str:
        if self._use_members:
            return cast(str, data.value)

        if self._lowercase_names:
            return data.name.lower()

        return data.name



"""
The guts of the library - the protocol state machine.
"""

from __future__ import annotations

import enum
import functools
import struct
from collections.abc import Callable, Collection, Mapping
from hashlib import md5
from itertools import count as it_count
from typing import Any, Self

import structlog
from scramp import ScramClient  # pyright: ignore[reportMissingTypeStubs]
from structlog.stdlib import BoundLogger

from pg_purepy.conversion import apply_default_converters
from pg_purepy.conversion.abc import ConversionContext, Converter
from pg_purepy.exc import (
    IllegalStateError,
    MissingPasswordError,
    ProtocolParseError,
    UnknownMessageError,
)
from pg_purepy.messages import (
    AuthenticationCompleted,
    AuthenticationMethod,
    AuthenticationRequest,
    BackendKeyData,
    BindComplete,
    ColumnDescription,
    CommandComplete,
    DataRow,
    ErrorOrNoticeResponse,
    ErrorResponseFieldType,
    ParameterDescription,
    ParameterStatus,
    ParseComplete,
    PortalSuspended,
    PostgresMessage,
    PreparedStatementInfo,
    ReadyForQuery,
    ReadyForQueryState,
    RowDescription,
    SASLComplete,
    SASLContinue,
)
from pg_purepy.util import Buffer, pack_strings

# static messages with no params
FLUSH_MESSAGE = b"H\x00\x00\x00\x04"
SYNC_MESSAGE = b"S\x00\x00\x00\x04"
TERMINATE_MESSAGE = b"X\x00\x00\x00\x04"
COPY_DONE_MESSAGE = b"c\x00\x00\x00\x04"
SSL_MESSAGE = b"\x00\x00\x00\x08\x04\xd2\x16/"


class BackendMessageCode(enum.IntEnum):
    """
    Enumeration of possible backend message codes.
    """

    NOTICE_RESPONSE = ord("N")
    AUTHENTICATION_REQUEST = ord("R")
    PARAMETER_STATUS = ord("S")
    BACKEND_KEY_DATA = ord("K")
    READY_FOR_QUERY = ord("Z")
    ROW_DESCRIPTION = ord("T")
    ERROR_RESPONSE = ord("E")
    DATA_ROW = ord("D")
    COMMAND_COMPLETE = ord("C")
    PARSE_COMPLETE = ord("1")
    BIND_COMPLETE = ord("2")
    CLOSE_COMPLETE = ord("3")
    PORTAL_SUSPENDED = ord("s")
    NO_DATA = ord("n")
    PARAMETER_DESCRIPTION = ord("t")
    NOTIFICATION_RESPONSE = ord("A")
    COPY_DONE = ord("c")
    COPY_DATA = ord("d")
    COPY_IN_RESPONSE = ord("G")
    COPY_OUT_RESPONSE = ord("H")
    EMPTY_QUERY_RESPONSE = ord("I")
    NEGOTIATE_PROTOCOL_VERSION = ord("v")


class FrontendMessageCode(enum.IntEnum):
    """
    Enumeration of possible frontend message codes.
    """

    BIND = ord("B")
    PARSE = ord("P")
    QUERY = ord("Q")
    EXECUTE = ord("E")
    FLUSH = ord("H")
    SYNC = ord("S")
    PASSWORD = ord("p")
    DESCRIBE = ord("D")
    TERMINATE = ord("X")
    CLOSE = ord("C")


class NeedData:
    """
    Special sentinel object returned to signify the state machine needs more data,
    """


#: Special singleton sentinel returned to signify the state machine needs more data.
NEED_DATA = NeedData()

#: Special singleton sentinel used to specify a state in the state machine cannot process any
#: events.
_NO_HANDLE = object()


def unrecoverable_error[Self_: "SansIOClient"](
    fn: Callable[[Self_, BackendMessageCode, Buffer], PostgresMessage],
) -> Callable[[Self_, BackendMessageCode, Buffer], PostgresMessage]:
    """
    Decorator that will automatically set the state to an unrecoverable error if an error response
    is found.
    """

    @functools.wraps(fn)
    def wrapper(
        self: Self_, code: BackendMessageCode, body: Buffer
    ) -> ErrorOrNoticeResponse | PostgresMessage:
        if code == BackendMessageCode.ERROR_RESPONSE:
            error = self._decode_error_response(body, recoverable=False, notice=False)
            if error.code == "57014":
                self.state = ProtocolState.RECOVERABLE_ERROR
                error.recoverable = True
            else:
                self.state = ProtocolState.UNRECOVERABLE_ERROR
                self._logger.critical(
                    "Unrecoverable error", severity=error.severity, message=error.message
                )

            return error

        return fn(self, code, body)

    return wrapper


class ProtocolState(enum.Enum):
    """
    Enumeration of possible states for the protocol machine.
    """

    ## Generic states.

    #: The initial state. We are waiting to send our startup message.
    STARTUP = 0

    #: We've sent our startup message. We're waiting for the server to respond to us.
    SENT_STARTUP = 1

    #: We've authenticated, receiving an Authentication Request Success, but we're waiting for
    #: a Ready For Query message.
    AUTHENTICATED_WAITING_FOR_COMPLETION = 2

    ## SASL states.

    #: During startup, we've been told to authenticate using SASL.
    SASL_STARTUP = 10

    #: During authentication, we've sent our first SASL message, and we're waiting for a response
    #: from the server.
    SASL_FIRST_SENT = 11

    #: During authentication, we've received the server's first SASL message, and we're waiting to
    #: send the final message.
    SASL_FIRST_RECEIVED = 12

    #: During authentication, we've sent our final SASL message, and we're waiting for a response
    #: from the server.
    SASL_FINAL_SENT = 13

    ## MD5 states.

    #: During startup, we've been told to authenticate using MD5.
    MD5_STARTUP = 20

    #: During authentication, we've sent our MD5-hashed password.
    MD5_SENT = 21

    ## Cleartext states.

    #: During startup, we've been told to authenticate using a cleartext password.
    CLEARTEXT_STARTUP = 30

    #: During authentication, we've sent our cleartext password.
    CLEARTEXT_SENT = 31

    ## Simple query states.

    #: During a simple query, we've sent the query to the server.
    SIMPLE_QUERY_SENT_QUERY = 100

    #: During a simple query, we've received the row description message, and we're waiting for
    #: either data rows, or command completion.
    SIMPLE_QUERY_RECEIVED_ROW_DESCRIPTION = 101

    #: During a simple query, we've received the command completion message, and we're waiting for
    #: the Ready for Query message.
    SIMPLE_QUERY_RECEIVED_COMMAND_COMPLETE = 102

    ## Extended query states.

    #: We've sent a Parse+Describe message to the server, and are waiting for a ParseComplete
    # message.
    MULTI_QUERY_SENT_PARSE_DESCRIBE = 300

    #: We've received a ParseComplete, and are waiting for the ParameterDescription+RowDescription
    #: messages.
    MULTI_QUERY_RECEIVED_PARSE_COMPLETE = 301

    #: We've received a ParameterDescription, and are waiting for a RowDescription. This state
    #: may be skipped.
    MULTI_QUERY_RECEIVED_PARAMETER_DESCRIPTION = 302

    #: We've received our RowDescription, and are waiting for the ReadyForQuery message.
    MULTI_QUERY_DESCRIBE_SYNC = 303

    #: We've sent a Bind message, and are waiting for a BindComplete.
    MULTI_QUERY_SENT_BIND = 310

    #: We've received a BindComplete or have unsuspended the portal, and are waiting for the data
    #: row messages.
    MULTI_QUERY_READING_DATA_ROWS = 311

    #: We've received a PortalSuspended message, and need to send another Execute.
    MULTI_QUERY_RECEIVED_PORTAL_SUSPENDED = 312

    #: We've received a CommandComplete messagee, and we are waiting for the ReadyForQuery
    #: message.
    MULTI_QUERY_RECEIVED_COMMAND_COMPLETE = 313

    ## Misc states.
    #: We're ready to send queries to the server. This is the state inbetween everything happening.
    READY_FOR_QUERY = 999

    #: An unrecoverable error has happened, and the protocol will no longer work.
    UNRECOVERABLE_ERROR = 1000

    #: A recoverable error has happened, and the protocol is waiting for a ReadyForQuery message.
    RECOVERABLE_ERROR = 1001

    #: The connection has been terminated.
    TERMINATED = 9999


class SansIOClient:
    """
    Sans-I/O state machine for the PostgreSQL C<->S protocol. This operates as an in-memory buffer
    that takes in Python-side structures and turns them into bytes to be sent to the server, and
    receives bytes from the server and turns them into Python-side structures.
    """

    _LOGGER_COUNTER = it_count()

    PROTOCOL_MAJOR = 3
    PROTOCOL_MINOR = 0

    _COMMANDS_WITH_COUNTS: Collection[str] = {"DELETE", "UPDATE", "SELECT", "MOVE", "FETCH", "COPY"}

    _PROTECTED_STATUSES: Mapping[str, str] = {
        "IntervalStyle": "iso_8601",
        "DateStyle": "ISO, DMY",
    }

    def __init__(
        self,
        username: str,
        database: str | None = None,
        password: str | None = None,
        application_name: str = "pg-purepy",
        logger_name: str | None = None,
        ignore_unknown_types: bool = True,
    ):
        """
        :param username: The username to authenticate as. Mandatory.
        :param database: The database to connect to. Defaults to ``username``.
        :param password: The password to authenticate with.
        :param application_name: The application name to send. Defaults to ``pg-purepy``.
        :param logger_name: The name of the logger to use. Defaults to a counter.
        :param ignore_unknown_types: If True, unknown types are returned as strings. Otherwise,
                                     raises an exception.
        """
        if not logger_name:
            logger_name = __name__ + f".protocol-{next(self._LOGGER_COUNTER)}"

        if database is None:
            database = username

        #: Mapping of miscellaneous connection parameters.
        self.connection_params: dict[str, Any] = {}

        #: If this connection is authenticated or not.
        self.is_authenticated: bool = False

        #: The converter classes used to convert between PostgreSQL and Python types.
        self.converters: dict[int, Converter[Any]] = {}

        #: Boolean value for if the current connection is in a transaction.
        self.in_transaction: bool = False

        # used for converters, and stores some parameter data so we dont have to update it in two
        # places.
        self._conversion_context = ConversionContext(client_encoding="UTF8")
        self._ignore_unknown_types = ignore_unknown_types

        # used when decoding data rows. unset by CommandComplete.
        self._last_row_description: RowDescription | None = None

        # used in two places:
        # 1) when a prepared statement is sent, we cache the name so we can return the ParseComplete
        # 2) when a describe is sent, we cache the name to be used in the resulting object
        self._last_named_query: str | None = None

        # used when a Describe message arrives.
        self._last_parameter_oids: list[int] = []

        # stored data used for generating various packets
        self.username = username
        self.database = database
        self.application_name: str = application_name
        self._password = password

        #: The state of the protocol state machine.
        self.state: ProtocolState = ProtocolState.STARTUP

        ## authentication
        # current authentication request. used by various auth functions
        self._auth_request: AuthenticationRequest | None = None
        # scramp client, used for SASL
        self._scramp_client: ScramClient | None = None

        ## buffering
        # infinitely growable buffer used to store incoming packets
        self._buffer = bytearray()
        # last packet code received
        self._current_packet_code = 0
        # number of bytes remaining in the current packet
        self._current_packet_remaining = 0
        # buffer used to store partial packet data
        self._current_packet_buffer = bytearray()
        # if the last receive started processing for a packet, but we didn't receive the full
        # data, we're currently processing a partial packet
        self._processing_partial_packet = False

        self._logger: BoundLogger = structlog.get_logger(name=__name__)
        self._logger = self._logger.bind(username=username, database=database)

        apply_default_converters(self)

    @property
    def ready(self) -> bool:
        """
        If the server is ready for another query.
        """
        return self.state == ProtocolState.READY_FOR_QUERY

    @property
    def encoding(self) -> str:
        """
        The client encoding.
        """

        return self._conversion_context.client_encoding

    @property
    def timezone(self) -> str:
        """
        The raw server timezone.
        """

        return self._conversion_context.timezone

    @property
    def has_pending_data(self) -> bool:
        """
        Checks if this has pending data within the buffer.
        """
        return len(self._buffer) > 0

    @property
    def dead(self) -> bool:
        """
        If the protocol is "dead", i.e. unusable via an error or terminated.
        """
        return self.state in {ProtocolState.UNRECOVERABLE_ERROR, ProtocolState.TERMINATED}

    ## Internal API ##
    def _check_password(self) -> None:
        if self._password is None:
            raise MissingPasswordError(
                "Server asked for authentication but we don't have a password"
            )

    def _handle_parameter_status(self, body: Buffer) -> ParameterStatus:
        """
        Handles a parameter status message. This will update the parameter statuses on the
        """
        encoding = self._conversion_context.client_encoding
        name, value = body.read_cstring(encoding), body.read_cstring(encoding)

        if name in self._PROTECTED_STATUSES and value != self._PROTECTED_STATUSES[name]:
            self.state = ProtocolState.UNRECOVERABLE_ERROR
            raise ProtocolParseError(f"Attempted to change protected connection parameter {name}!")

        if name == "client_encoding":
            self._conversion_context.client_encoding = value
        elif name == "TimeZone":
            self._conversion_context.timezone = value
        else:
            self.connection_params[name] = value

        self._logger.debug("Set parameter status", parameter=name, value=value)

        return ParameterStatus(name, value)

    def _decode_error_response(
        self,
        body: Buffer,
        *,
        recoverable: bool,
        notice: bool,
    ) -> ErrorOrNoticeResponse:
        """
        Decodes an error response, returning the :class:`.ErrorOrNoticeResponse` containing all the
        fields from the message.
        """
        kwargs: dict[str, Any] = {"recoverable": recoverable, "notice": notice}
        ignored_codes: set[int] = {ord("F"), ord("L"), ord("R")}

        while body:
            try:
                raw_code = body.read_byte()
                if raw_code == 0:  # final null sep
                    break

                if raw_code in ignored_codes:
                    # discard body
                    body.read_cstring(encoding="ascii")
                    continue

                code = ErrorResponseFieldType(raw_code)
            except ValueError:
                code = ErrorResponseFieldType.UNKNOWN

            str_field = body.read_cstring(encoding="ascii")
            if code == ErrorResponseFieldType.UNKNOWN:
                self._logger.warning("Unknown field type in error", field=str_field)
            else:
                kwargs[code.name.lower()] = str_field

        return ErrorOrNoticeResponse(**kwargs)

    def _decode_row_description(self, body: Buffer) -> RowDescription:
        """
        Decodes the row description message.
        """
        field_count = body.read_short()
        fields: list[ColumnDescription] = []

        for _i in range(field_count):
            name = body.read_cstring(encoding=self.encoding)
            table_oid = body.read_int()
            column_idx = body.read_short()
            type_oid = body.read_int()

            if type_oid not in self.converters:
                if self._ignore_unknown_types:
                    self._logger.warning("Unknown type OID", oid=type_oid)
                else:
                    raise ProtocolParseError(f"Unknown type OID: {type_oid}")

            type_size = body.read_short()
            type_mod = body.read_int()
            format_code = body.read_short()  # noqa: F841, good for code explicitness

            desc = ColumnDescription(
                name=name,
                table_oid=table_oid,
                column_index=column_idx,
                type_oid=type_oid,
                column_length=type_size,
                type_modifier=type_mod,
            )
            fields.append(desc)

        return RowDescription(fields)

    def _decode_data_row(self, body: Buffer) -> DataRow:
        """
        Decodes a single data row packet.
        """
        assert self._last_row_description is not None, "missing row description"

        column_values: list[Any | None] = []
        count = body.read_short()

        for idx in range(count):
            size = body.read_int()

            # -1 size means null.
            if size < 0:
                column_values.append(None)
                continue

            data = body.read_bytes(size)

            col_desc = self._last_row_description.columns[idx]
            col_oid = col_desc.type_oid
            converter = self.converters.get(col_oid)
            if converter is None:
                column_values.append(data.decode(self.encoding))
            else:
                column_values.append(
                    converter.from_postgres(self._conversion_context, data.decode(self.encoding))
                )

        return DataRow(self._last_row_description, column_values)

    def _decode_command_complete(self, body: Buffer) -> CommandComplete:
        """
        Decodes a "command complete" message.
        """
        tag = body.read_cstring(encoding=self.encoding)
        split = tag.split(" ")

        if split[0] == "INSERT":
            row_count = int(split[2])
        elif split[0] in self._COMMANDS_WITH_COUNTS:
            row_count = int(split[1])
        else:
            row_count = None

        return CommandComplete(split[0], row_count)

    _handle_during_STARTUP = "You need to call do_startup()!"

    ### Authentication ###
    @unrecoverable_error
    def _handle_during_SENT_STARTUP(
        self, code: BackendMessageCode, body: Buffer
    ) -> AuthenticationCompleted | AuthenticationRequest:
        if code == BackendMessageCode.NEGOTIATE_PROTOCOL_VERSION:
            raise ProtocolParseError("Server does not support our protocol")

        if code == BackendMessageCode.AUTHENTICATION_REQUEST:
            # various different options
            decoded = body.read_int()

            # AuthenticationOk. We skip authentication entirely.
            if decoded == 0:
                self.is_authenticated = True
                self.state = ProtocolState.AUTHENTICATED_WAITING_FOR_COMPLETION
                return AuthenticationCompleted()

            # AuthenticationCleartextPassword
            if decoded == AuthenticationMethod.CLEARTEXT:
                self._check_password()

                self.state = ProtocolState.CLEARTEXT_STARTUP
                self._auth_request = AuthenticationRequest(method=AuthenticationMethod.CLEARTEXT)
                return self._auth_request

            # AuthenticationMD5
            if decoded == AuthenticationMethod.MD5:
                self._check_password()

                self.state = ProtocolState.MD5_STARTUP
                md5_bytes = body.read_bytes(4)
                self._auth_request = AuthenticationRequest(
                    method=AuthenticationMethod.MD5, md5_salt=md5_bytes
                )
                return self._auth_request

            # AuthenticationSASL
            if decoded == AuthenticationMethod.SASL:
                self._check_password()

                self.state = ProtocolState.SASL_STARTUP
                methods = body.read_all_cstrings(encoding=self.encoding, drop_empty=True)
                self._auth_request = AuthenticationRequest(
                    method=AuthenticationMethod.SASL,
                    sasl_methods=methods,
                )
                return self._auth_request

            raise NotImplementedError(f"Unknown authentication method: {decoded}")

        raise UnknownMessageError(f"Don't know how to handle {code} in {self.state}")

    def _handle_during_SASL_FIRST_SENT(
        self, code: BackendMessageCode, body: Buffer
    ) -> SASLContinue:
        if code == BackendMessageCode.AUTHENTICATION_REQUEST:
            body_code = body.read_int()
            if body_code == 11:  # SASL Continue
                sasl_body = body.read_remaining().decode("ascii")

                assert self._scramp_client
                self._scramp_client.set_server_first(sasl_body)  # type: ignore
                self.state = ProtocolState.SASL_FIRST_RECEIVED
                return SASLContinue()

            raise ProtocolParseError(
                f"Unexpected response: Expected SASLContinue, but got {body_code}"
            )

        raise UnknownMessageError(f"Unknown message code {code} for state {self.state}")

    @unrecoverable_error
    def _generic_auth_handle(
        self, code: BackendMessageCode, body: Buffer
    ) -> AuthenticationCompleted | SASLComplete:
        """
        Authentication handler shared between various authentication "finalisation" methods.
        """
        if code == BackendMessageCode.AUTHENTICATION_REQUEST:
            body_code = body.read_int()
            if body_code == 0:  # AuthenticationOk
                self.is_authenticated = True
                self.state = ProtocolState.AUTHENTICATED_WAITING_FOR_COMPLETION
                return AuthenticationCompleted()

            if body_code == 12:  # SASLComplete
                return SASLComplete()

            raise ProtocolParseError(
                f"Unexpected response: Expected AuthenticationOk, but got {body_code}"
            )

        raise UnknownMessageError(f"Unknown message code {code} for state {self.state}")

    _handle_during_CLEARTEXT_SENT = _generic_auth_handle
    _handle_during_MD5_SENT = _generic_auth_handle
    _handle_during_SASL_FINAL_SENT = _generic_auth_handle

    def _got_ready_for_query(self, body: Buffer) -> ReadyForQuery:
        """
        Helper function used when a ReadyForQuery is received.
        """
        self.state = ProtocolState.READY_FOR_QUERY
        rfqs = ReadyForQueryState(body.read_byte())
        self.in_transaction = rfqs != ReadyForQueryState.IDLE

        self._last_row_description = None  # common reset

        return ReadyForQuery(rfqs)

    @unrecoverable_error
    def _handle_ready_for_query(self, code: BackendMessageCode, body: Buffer) -> ReadyForQuery:
        """
        Generic handler for when the server is ready for query.
        """
        if code == BackendMessageCode.READY_FOR_QUERY:
            return self._got_ready_for_query(body)

        raise UnknownMessageError(f"Expected ReadyForQuery, got {code!r}")

    @unrecoverable_error
    def _handle_during_AUTHENTICATED_WAITING_FOR_COMPLETION(
        self,
        code: BackendMessageCode,
        body: Buffer,
    ) -> ReadyForQuery | BackendKeyData:
        """
        Handles incoming messages once we've authenticated.
        """
        if code == BackendMessageCode.READY_FOR_QUERY:
            return self._got_ready_for_query(body)

        if code == BackendMessageCode.BACKEND_KEY_DATA:
            pid = body.read_int()
            key = body.read_uint()
            return BackendKeyData(pid, key)

        raise UnknownMessageError(f"Expected ReadyForQuery, got {code!r}")

    ### Simple Query ###
    def _got_row_description(self, body: Buffer) -> RowDescription:
        """
        Common logic for when we get a row description.
        """
        row_desc = self._decode_row_description(body)
        self.state = ProtocolState.SIMPLE_QUERY_RECEIVED_ROW_DESCRIPTION
        self._last_row_description = row_desc
        return row_desc

    def _handle_during_SIMPLE_QUERY_SENT_QUERY(
        self, code: BackendMessageCode, body: Buffer
    ) -> CommandComplete | RowDescription | ErrorOrNoticeResponse:
        """
        Handles incoming messages once we've sent a query execution message.
        """
        if code == BackendMessageCode.COMMAND_COMPLETE:
            # commands such as Begin or Rollback don't return rows, thus don't return a
            # RowDescription.
            self.state = ProtocolState.SIMPLE_QUERY_RECEIVED_COMMAND_COMPLETE
            return self._decode_command_complete(body)

        if code == BackendMessageCode.ROW_DESCRIPTION:
            return self._got_row_description(body)

        if code == BackendMessageCode.ERROR_RESPONSE:
            # this is a recoverable error, and simply moves onto the ReadyForQuery message.
            self.state = ProtocolState.RECOVERABLE_ERROR
            error = self._decode_error_response(body, recoverable=True, notice=False)
            self._logger.warning(
                "Recoverable error during query", severity=error.severity, message=error.message
            )
            return error

        raise UnknownMessageError(f"Expected RowDescription, got {code!r}")

    @unrecoverable_error
    def _handle_during_SIMPLE_QUERY_RECEIVED_ROW_DESCRIPTION(
        self, code: BackendMessageCode, body: Buffer
    ) -> DataRow | CommandComplete:
        """
        Handles incoming data rows during a simple query.
        """
        if code == BackendMessageCode.DATA_ROW:
            return self._decode_data_row(body)

        if code == BackendMessageCode.COMMAND_COMPLETE:
            self.state = ProtocolState.SIMPLE_QUERY_RECEIVED_COMMAND_COMPLETE
            self._last_row_description = None
            return self._decode_command_complete(body)

        raise UnknownMessageError(f"Expected DataRow or CommandComplete, got {code!r}")

    @unrecoverable_error
    def _handle_during_SIMPLE_QUERY_RECEIVED_COMMAND_COMPLETE(
        self, code: BackendMessageCode, body: Buffer
    ) -> RowDescription | ReadyForQuery:
        """
        Waits for the ReadyForQuery response, or the next row description if this was a multi-query.
        """
        if code == BackendMessageCode.ROW_DESCRIPTION:
            # multiple queries in one statement
            return self._got_row_description(body)

        if code == BackendMessageCode.READY_FOR_QUERY:
            return self._got_ready_for_query(body)

        raise UnknownMessageError(f"Unexpected message {code!r}")

    _handle_during_SIMPLE_QUERY_ERRORED: Callable[
        [Self, BackendMessageCode, Buffer], PostgresMessage
    ] = _handle_ready_for_query

    ### Multi query, Prepared Statements ###
    def _handle_during_MULTI_QUERY_SENT_PARSE_DESCRIBE(
        self, code: BackendMessageCode, body: Buffer
    ) -> ParseComplete | ErrorOrNoticeResponse:
        """
        Waits for the ParseComplete response.
        """
        if code == BackendMessageCode.PARSE_COMPLETE:
            self.state = ProtocolState.MULTI_QUERY_RECEIVED_PARSE_COMPLETE
            return ParseComplete(self._last_named_query)

        if code == BackendMessageCode.ERROR_RESPONSE:
            # bad query
            self.state = ProtocolState.RECOVERABLE_ERROR
            return self._decode_error_response(body, recoverable=True, notice=False)

        raise UnknownMessageError(f"Expected ParseComplete, got {code!r}")

    def _multi_query_got_description(self, body: Buffer) -> PreparedStatementInfo:
        decoded = self._decode_row_description(body)
        lnq = self._last_named_query if self._last_named_query else None
        self._last_named_query = None
        message = PreparedStatementInfo(
            lnq, ParameterDescription(self._last_parameter_oids), decoded
        )
        self._last_parameter_oids = []

        self.state = ProtocolState.MULTI_QUERY_DESCRIBE_SYNC

        return message

    def _multi_query_got_no_data(self) -> PreparedStatementInfo:
        lnq = self._last_named_query if self._last_named_query else None
        self._last_named_query = None
        message = PreparedStatementInfo(lnq, ParameterDescription(self._last_parameter_oids), None)

        self.state = ProtocolState.MULTI_QUERY_DESCRIBE_SYNC
        return message

    @unrecoverable_error
    def _handle_during_MULTI_QUERY_RECEIVED_PARSE_COMPLETE(
        self, code: BackendMessageCode, body: Buffer
    ) -> ParameterDescription | PreparedStatementInfo:
        """
        Decodes a possible ParameterDescription message, or a RowDescription message.
        """
        if code == BackendMessageCode.PARAMETER_DESCRIPTION:
            count = body.read_short()
            oids = [body.read_int() for _ in range(count)]
            self._last_parameter_oids = oids
            self.state = ProtocolState.MULTI_QUERY_RECEIVED_PARAMETER_DESCRIPTION
            return ParameterDescription(oids)

        if code == BackendMessageCode.ROW_DESCRIPTION:
            return self._multi_query_got_description(body)

        raise UnknownMessageError(f"Expected ParameterDescription or RowDescription, got {code!r}")

    @unrecoverable_error
    def _handle_during_MULTI_QUERY_RECEIVED_PARAMETER_DESCRIPTION(
        self, code: BackendMessageCode, body: Buffer
    ) -> PreparedStatementInfo:
        """
        Decodes the incoming row description message.
        """
        # Both of these change state to MULTI_QUERY_DESCRIBE_SYNC
        if code == BackendMessageCode.ROW_DESCRIPTION:
            # Changes state to MULTI_QUERY_DESCRIBE_SYNC
            return self._multi_query_got_description(body)

        if code == BackendMessageCode.NO_DATA:
            return self._multi_query_got_no_data()

        raise UnknownMessageError(f"Expected RowDescription or NoData, got {code!r}")

    _handle_during_MULTI_QUERY_DESCRIBE_SYNC = _handle_ready_for_query

    ### Multi query, incoming data ###
    def _handle_during_MULTI_QUERY_SENT_BIND(
        self, code: BackendMessageCode, body: Buffer
    ) -> BindComplete | ErrorOrNoticeResponse:
        """
        Waits for the BindComplete message.
        """
        if code == BackendMessageCode.BIND_COMPLETE:
            self.state = ProtocolState.MULTI_QUERY_READING_DATA_ROWS
            return BindComplete()

        if code == BackendMessageCode.ERROR_RESPONSE:
            self.state = ProtocolState.RECOVERABLE_ERROR
            return self._decode_error_response(body, recoverable=True, notice=False)

        raise UnknownMessageError(f"Expected BindComplete, got {code!r}")

    def _handle_during_MULTI_QUERY_READING_DATA_ROWS(
        self, code: BackendMessageCode, body: Buffer
    ) -> DataRow | CommandComplete | ErrorOrNoticeResponse | PortalSuspended:
        """
        Reads in data rows, and waits for CommandComplete and ReadyForQuery.
        """
        if code == BackendMessageCode.DATA_ROW:
            return self._decode_data_row(body)

        if code == BackendMessageCode.COMMAND_COMPLETE:
            self.state = ProtocolState.MULTI_QUERY_RECEIVED_COMMAND_COMPLETE
            return self._decode_command_complete(body)

        if code == BackendMessageCode.ERROR_RESPONSE:
            # TODO: Maybe check the code, and reraise as unrecoverable?
            self.state = ProtocolState.RECOVERABLE_ERROR
            return self._decode_error_response(body, recoverable=True, notice=False)

        if code == BackendMessageCode.PORTAL_SUSPENDED:
            self.state = ProtocolState.MULTI_QUERY_RECEIVED_PORTAL_SUSPENDED
            return PortalSuspended()

        raise UnknownMessageError(f"Expected DataRow or CommandComplete, got {code!r}")

    # The docs seem to lie here, the server just sends us the PortalSuspended then
    # ReadyForQuery. I don't get it.
    # pg8000 seems to just drop it on the floor too... so that's what we do.
    _handle_during_MULTI_QUERY_RECEIVED_PORTAL_SUSPENDED = _handle_ready_for_query
    _handle_during_MULTI_QUERY_RECEIVED_COMMAND_COMPLETE = _handle_ready_for_query
    _handle_during_RECOVERABLE_ERROR = _handle_ready_for_query

    ## Public API ##
    def add_converter(self, converter: Converter[Any]) -> None:
        """
        Adds a new :class:`.Converter` to this protocol.
        """
        self.converters[converter.oid] = converter

    def receive_bytes(self, data: bytes) -> None:
        """
        Receives incoming bytes from the server. This merely appends to an internal buffer; you
        need to call :meth:`.next_event` to do anything.

        :param data: The incoming data from PostgreSQL.
        """
        # these are assertions so that ``python -O`` optimises them out. its a bug to do anything
        # once the protocol is in UNRECOVERABLE_ERROR anyway.
        assert self.state != ProtocolState.UNRECOVERABLE_ERROR, "state is unrecoverable error"
        assert self.state != ProtocolState.TERMINATED, "state is terminated"

        self._buffer += data

    def do_startup(self) -> bytes:
        """
        Gets the startup packet, and sets the state machine up for starting up.

        :return: A startup message, that should be sent to PostgreSQL.
        """
        if self.state != ProtocolState.STARTUP:
            raise IllegalStateError("Can't do startup outside of the startup state")

        packet_body = struct.pack("!i", (self.PROTOCOL_MAJOR << 16) | self.PROTOCOL_MINOR)
        packet_body += pack_strings("user", self.username)
        packet_body += pack_strings("database", self.database)
        packet_body += pack_strings("application_name", self.application_name)
        packet_body += pack_strings("client_encoding", self._conversion_context.client_encoding)
        packet_body += pack_strings("IntervalStyle", "iso_8601")
        packet_body += pack_strings("DateStyle", "ISO, DMY")
        packet_body += b"\x00"
        size = struct.pack("!i", (len(packet_body) + 4))

        self.state = ProtocolState.SENT_STARTUP
        return size + packet_body

    def do_terminate(self) -> bytes:
        """
        Gets the Terminate message. This closes the protocol!
        """
        self.state = ProtocolState.TERMINATED
        return TERMINATE_MESSAGE

    def do_simple_query(self, query_text: str) -> bytes:
        """
        Performs a simple (static) query. This query cannot have any dynamic parameters.
        """
        if self.state != ProtocolState.READY_FOR_QUERY:
            raise IllegalStateError("The server is not ready for queries")

        packet_body = pack_strings(query_text, encoding=self.encoding)
        header = FrontendMessageCode.QUERY.to_bytes(length=1, byteorder="big")
        header += (len(packet_body) + 4).to_bytes(length=4, byteorder="big")

        self.state = ProtocolState.SIMPLE_QUERY_SENT_QUERY
        return header + packet_body

    def do_create_prepared_statement(self, name: str | None, query_text: str) -> bytes:
        """
        Creates a prepared statement. If ``name`` is not specified, this will create the unnamed
        prepared statement.
        """
        # TIL: You can do Parse, then Describe, then Flush, and get all the messages back at once.
        # This saves a Flush and a Sync round trip.

        if self.state != ProtocolState.READY_FOR_QUERY:
            raise IllegalStateError("The server is not ready for queries")

        ## Parse
        if name is None:
            packet_body_1 = b"\x00"  # empty string
            self._last_named_query = None
        else:
            packet_body_1 = pack_strings(name, encoding="ascii")
            self._last_named_query = name

        packet_body_1 += pack_strings(query_text, encoding=self.encoding)
        # no OIDs. (yet). probably in the future...
        packet_body_1 += b"\x00\x00"

        header_1 = FrontendMessageCode.PARSE.to_bytes(length=1, byteorder="big")
        header_1 += (len(packet_body_1) + 4).to_bytes(length=4, byteorder="big")
        full_packet = header_1 + packet_body_1

        ## Describe
        packet_body_2 = b"S"

        if name:
            packet_body_2 += pack_strings(name, encoding="ascii")
        else:
            packet_body_2 += b"\x00"

        header_2 = FrontendMessageCode.DESCRIBE.to_bytes(length=1, byteorder="big")
        header_2 += (len(packet_body_2) + 4).to_bytes(length=4, byteorder="big")
        full_packet += header_2 + packet_body_2
        full_packet += FLUSH_MESSAGE
        full_packet += SYNC_MESSAGE

        self.state = ProtocolState.MULTI_QUERY_SENT_PARSE_DESCRIBE
        return full_packet

    def do_bind_execute(
        self, info: PreparedStatementInfo, params: Any, row_count: int | None = None
    ) -> bytes:
        """
        Binds the specified ``params`` to the prepared statement specified by ``info``,
        then executes it.

        If ``row_count`` is not None, then only that many rows will be returned at maximum.
        Otherwise, an unlimited amount of rows will be returned.

        :returns: The combined bind-execute-flush-sync packet to send to the server.
        """
        wanted_params = len(info.parameter_oids.oids)
        if len(params) != wanted_params:
            raise ValueError(f"Expected exactly {wanted_params} parameters, got {len(params)}")

        ## Packet: Bind

        ### Bind: Portal Name
        # Unsupported. Empty.
        packet_body_1: bytes = pack_strings("")

        ### Bind: Source Prepared Statement
        if info.name is None:
            packet_body_1 += b"\x00"
        else:
            packet_body_1 += pack_strings(info.name, encoding="ascii")

        ### Bind: Number of parameter format codes.
        packet_body_1 += struct.pack(">h", wanted_params)
        ### Bind: Parameter format codes.
        for _ in range(wanted_params):
            packet_body_1 += struct.pack(">h", 0)

        ### Bind: Number of parameter values.
        packet_body_1 += struct.pack(">h", wanted_params)
        ### Bind: Parameter values.
        for oid, obb in zip(info.parameter_oids.oids, params, strict=True):
            if obb is None:
                # Nulls are special cased with a -1 length.
                packet_body_1 += struct.pack(">i", -1)
                continue

            try:
                converter = self.converters[oid]
            except KeyError:
                raise ValueError(f"Missing converter for {oid}, can't convert {obb}") from None

            converted = converter.to_postgres(self._conversion_context, obb)
            assert isinstance(converted, str), f"Converter returned {type(obb)}, not string!"

            encoded = converted.encode(self.encoding)
            packet_body_1 += struct.pack(">i", len(encoded))
            packet_body_1 += encoded

        ### Bind: Result columns formats.
        wanted_results = len(info.row_description.columns) if info.row_description else 0

        packet_body_1 += struct.pack(">h", wanted_results)
        for _ in range(wanted_results):
            packet_body_1 += struct.pack(">h", 0)

        full_packet = FrontendMessageCode.BIND.to_bytes(length=1, byteorder="big")
        full_packet += (len(packet_body_1) + 4).to_bytes(length=4, byteorder="big")
        full_packet += packet_body_1

        ## Execute packet
        ### Execute: Unnamed portal.
        packet_body_2 = pack_strings()
        row_count = row_count or 0

        ### Execute: Row count.
        packet_body_2 += struct.pack(">i", row_count)

        full_packet += FrontendMessageCode.EXECUTE.to_bytes(length=1, byteorder="big")
        full_packet += (len(packet_body_2) + 4).to_bytes(length=4, byteorder="big")
        full_packet += packet_body_2

        ## These messages include a Flush + Sync.
        full_packet += FLUSH_MESSAGE
        full_packet += SYNC_MESSAGE

        self.state = ProtocolState.MULTI_QUERY_SENT_BIND
        self._last_row_description = info.row_description
        return full_packet

    def next_event(self) -> PostgresMessage | NeedData:
        """
        Reads the next event from the message queue.

        :return: Either a :class:`.PostgresMessage`, or the special ``NEED_DATA`` constant.
        """
        if (
            self.state == ProtocolState.UNRECOVERABLE_ERROR
            or self.state == ProtocolState.TERMINATED
        ):
            raise IllegalStateError("The protocol is broken and won't work.")

        if len(self._buffer) == 0:
            return NEED_DATA

        # This hurts my brain

        if self._processing_partial_packet:
            # rough flow:
            # 1) copy off the main buffer into the packet buffer
            # 2) if we have enough data, unflip the flag, reset buffers
            # 3) if we don't have enough data, just return NEED_DATA, we'll check again next loop

            remaining = self._current_packet_remaining
            assert remaining > 0, "partial packet needs actual data to read"
            data, self._buffer = self._buffer[0:remaining], self._buffer[remaining:]
            self._current_packet_buffer += data

            self._current_packet_remaining -= len(data)
            if self._current_packet_remaining > 0:
                return NEED_DATA

            self._processing_partial_packet = False
            code = self._current_packet_code
            message_data = self._current_packet_buffer
            self._current_packet_buffer = bytearray()
        else:
            # no code + size, don't care
            if len(self._buffer) < 5:
                return NEED_DATA

            code = self._buffer[0]
            size = int.from_bytes(self._buffer[1:5], byteorder="big") - 4
            message_data = self._buffer[5 : size + 5]

            if len(message_data) < size:
                # not enough data yet.
                self._processing_partial_packet = True
                self._current_packet_buffer = message_data
                self._current_packet_code = code
                self._current_packet_remaining = size - len(message_data)
                # reset buffer, otherwise we try and read the message code off again >.>
                self._buffer = bytearray()
                return NEED_DATA
            else:  # noqa: RET505
                # yes enough data, set the buffer to the data after the size bytes for future
                # processing
                self._buffer = self._buffer[size + 5 :]

        code = BackendMessageCode(code)
        self._logger.debug("Incoming protocol message", code=code)

        try:
            method = getattr(self, f"_handle_during_{self.state.name}")
        except AttributeError:
            raise AttributeError(f"No method exists to process state {self.state.name}") from None

        if isinstance(method, str):
            raise IllegalStateError(method)

        buffer = Buffer(message_data)

        # These messages can happen at any state.
        # As such, we yield them out unconditionally. Typically, a client ignores them and carries
        # on reading messages off.

        if code == BackendMessageCode.PARAMETER_STATUS:
            return self._handle_parameter_status(buffer)

        if code == BackendMessageCode.NOTICE_RESPONSE:
            return self._decode_error_response(body=buffer, recoverable=True, notice=True)

        # if we made it here, we can actually process the packet since we have it in full

        result: PostgresMessage | ErrorOrNoticeResponse = method(code, buffer)

        if result is None:  # type: ignore
            raise Exception(f"{method.__qualname__} returned None! This is a bug!")

        return result

    def get_needed_synchronisation(self) -> bytes:
        """
        Gets the bytes to send to the server if we need to do synchronisation. This is used e.g.
        during a complex query message where a Sync packet is needed, or authentication where
        multiple messages may need to be sent.

        This may change the state of the protocol machine.
        """
        packet_body = b""

        # Blank states are marked eexplicitly for the sake of ease of reading.

        ## Authentication States ##
        if self.state == ProtocolState.CLEARTEXT_STARTUP:
            self._check_password()

            code = FrontendMessageCode.PASSWORD

            # type ignore is fine, ``_check_password`` verifies itt.
            assert self._password is not None
            packet_body += self._password.encode(encoding="ascii")
            self.state = ProtocolState.CLEARTEXT_SENT

        elif self.state == ProtocolState.MD5_STARTUP:
            assert self._auth_request, "state is MD5_STARTUP but no auth request packet?"
            salt = self._auth_request.md5_salt
            assert salt, "state is MD5_STARTUP but salt is None?"
            self._check_password()

            assert self._password is not None

            code = FrontendMessageCode.PASSWORD
            hashed_password = md5(self._password.encode("ascii") + self.username.encode("ascii"))
            inner_password = hashed_password.hexdigest().encode("ascii")
            encoded_password = md5(inner_password + salt).hexdigest().encode("ascii")
            packet_body += b"md5"
            packet_body += encoded_password
            packet_body += b"\x00"

            self.state = ProtocolState.MD5_SENT

        elif self.state == ProtocolState.SASL_STARTUP:
            code = FrontendMessageCode.PASSWORD
            assert self._auth_request, "state is SASL_STARTUP but no auth request packet?"

            self._scramp_client = ScramClient(
                mechanisms=self._auth_request.sasl_methods,
                username=self.username,
                password=self._password,
            )

            message = self._scramp_client.get_client_first()
            packet_body += pack_strings(self._scramp_client.mechanism_name)
            packet_body += len(message).to_bytes(length=4, byteorder="big", signed=False)
            packet_body += message.encode("ascii")

            self.state = ProtocolState.SASL_FIRST_SENT

        elif self.state == ProtocolState.SASL_FIRST_RECEIVED:
            assert self._scramp_client, "state is SASL_FIRST_RECEIVED but scramp client is null?"

            code = FrontendMessageCode.PASSWORD
            message = self._scramp_client.get_client_final()
            packet_body += message.encode("ascii")
            self.state = ProtocolState.SASL_FINAL_SENT

        else:
            return b""

        msg_len = (len(packet_body) + 4).to_bytes(length=4, byteorder="big", signed=False)
        return code.to_bytes(length=1, byteorder="big") + msg_len + packet_body


# TLS helpers
def check_if_tls_accepted(reply: bytes) -> bool:
    """
    Checks if the server accepted the TLS request.
    """
    return reply == b"S"



from enum import Enum, StrEnum

import pytest
from pg_purepy.conversion.enums import EnumConverter

from tests.util import open_connection

pytestmark = pytest.mark.anyio


class StandardEnum(Enum):
    ONE = 1
    TWO = 2
    THREE = 3


class StringEnum(StrEnum):
    THREE = "one"
    ONE = "two"
    TWO = "three"


async def test_basic_enum_conversion():
    """
    Tests basic enum conversion.
    """

    async with open_connection() as conn:
        await conn.execute("drop type if exists test_bec;")
        await conn.execute("create type test_bec as enum ('one', 'two', 'three');")
        await conn.execute("create temp table test_bec_t (id test_bec primary key);")

        oid_row = await conn.fetch_one("select oid from pg_type where typname = 'test_bec';")
        assert oid_row
        oid = oid_row.data[0]
        assert oid
        converter = EnumConverter(oid, StandardEnum)
        conn.add_converter(converter)

        # ok now for the actual fun stuff
        await conn.execute("insert into test_bec_t(id) values (:one);", one=StandardEnum.ONE)
        data_row = await conn.fetch_one("select id from test_bec_t;")
        assert data_row
        assert data_row.data[0] == StandardEnum.ONE


async def test_string_member_enum_conversion():
    """
    Tests enum conversion using string members.
    """

    async with open_connection() as conn:
        await conn.execute("drop type if exists test_smec;")
        await conn.execute("create type test_smec as enum ('one', 'two', 'three');")
        await conn.execute("create temp table test_smec_t (id test_smec primary key);")

        oid_row = await conn.fetch_one("select oid from pg_type where typname = 'test_smec';")
        assert oid_row
        oid = oid_row.data[0]
        assert oid
        converter = EnumConverter(oid, StringEnum, use_member_values=True)
        conn.add_converter(converter)

        await conn.execute("insert into test_smec_t(id) values ('one');")
        data_row = await conn.fetch_one("select id from test_smec_t;")
        assert data_row
        assert data_row.data[0] == StringEnum.THREE



## The below function is taken and modified from the SQLAlchemy project which is licensed under the
## MIT License.
##
## Copyright 2005-2021 SQLAlchemy authors and contributors <see AUTHORS file>.
##
## Permission is hereby granted, free of charge, to any person obtaining a copy of
## this software and associated documentation files (the "Software"), to deal in
## the Software without restriction, including without limitation the rights to
## use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
## of the Software, and to permit persons to whom the Software is furnished to do
## so, subject to the following conditions:
##
## The above copyright notice and this permission notice shall be included in all
## copies or substantial portions of the Software.
##
## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
## IMPLIEkD, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
## FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
## AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
## LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
## OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
## SOFTWARE.

# mypy: ignore-errors
# pyright: basic

from __future__ import annotations

import re
from collections.abc import Mapping

# My best guess at the parsing rules of hstore literals, since no formal
# grammar is given.  This is mostly reverse engineered from PG's input parser
# behavior.


HSTORE_PAIR_RE = re.compile(
    r"""
(
  "(?P<key> (\\ . | [^"])* )"       # Quoted key
)
[ ]* => [ ]*    # Pair operator, optional adjoining whitespace
(
    (?P<value_null> NULL )          # NULL value
  | "(?P<value> (\\ . | [^"])* )"   # Quoted value
)
""",
    re.VERBOSE,
)

HSTORE_DELIMITER_RE = re.compile(
    r"""
[ ]* , [ ]*
""",
    re.VERBOSE,
)


def _parse_hstore(hstore_str: str) -> Mapping[str, str]:
    """
    Parse an hstore from its literal string representation.

    Attempts to approximate PG's hstore input parsing rules as closely as
    possible. Although currently this is not strictly necessary, since the
    current implementation of hstore's output syntax is stricter than what it
    accepts as input, the documentation makes no guarantees that will always
    be the case.
    """

    result = {}
    pos = 0
    pair_match = HSTORE_PAIR_RE.match(hstore_str)

    while pair_match is not None:
        key = pair_match.group("key").replace(r"\"", '"').replace("\\\\", "\\")
        if pair_match.group("value_null"):
            value = None
        else:
            value = pair_match.group("value").replace(r"\"", '"').replace("\\\\", "\\")
        result[key] = value

        pos += pair_match.end()

        delim_match = HSTORE_DELIMITER_RE.match(hstore_str[pos:])
        if delim_match is not None:
            pos += delim_match.end()

        pair_match = HSTORE_PAIR_RE.match(hstore_str[pos:])

    if pos != len(hstore_str):
        raise ValueError(f"failed to parse {hstore_str} at {pos}")

    return result


def _serialize_hstore(val: Mapping[str, str]) -> str:
    """Serialize a dictionary into an hstore literal.  Keys and values must
    both be strings (except None for values).
    """

    def esc(s, position):
        if position == "value" and s is None:
            return "NULL"

        if isinstance(s, str):
            return '"{}"'.format(s.replace("\\", "\\\\").replace('"', r"\""))

        raise ValueError(f"{s!r} in {position} position is not a string.")

    return ", ".join("{}=>{}".format(esc(k, "key"), esc(v, "value")) for k, v in val.items())



from __future__ import annotations

from collections import deque


def pack_strings(*s: str, encoding: str = "ascii") -> bytes:
    """
    Packs a sequence of strings using null terminators.
    """
    return b"\x00".join(x.encode(encoding) for x in s) + b"\x00"


class Buffer:
    """
    Simple buffer that allows reading data off of a bytearray ala Java ByteBuffer.
    """

    def __init__(self, ba: bytearray | None = None) -> None:
        self.data: deque[int] = deque(ba if ba else b"")

    def __bool__(self) -> bool:
        return bool(self.data)

    def __len__(self) -> int:
        return len(self.data)

    def read_bytes(self, count: int) -> bytearray:
        ba = bytearray()
        for _x in range(count):
            ba.append(self.data.popleft())

        return ba

    def read_byte(self) -> int:
        return self.data.popleft()

    def read_short(self) -> int:
        return int.from_bytes(self.read_bytes(2), byteorder="big", signed=True)

    def read_int(self) -> int:
        return int.from_bytes(self.read_bytes(4), byteorder="big", signed=True)

    def read_uint(self) -> int:
        return int.from_bytes(self.read_bytes(4), byteorder="big", signed=False)

    def read_long(self) -> int:
        return int.from_bytes(self.read_bytes(8), byteorder="big", signed=True)

    def read_cstring(self, encoding: str) -> str:
        """
        Reads a null-terminated C string.
        """
        buf = bytearray()
        while True:
            byte = self.data.popleft()
            if byte == 0x0:
                break

            buf.append(byte)

        return buf.decode(encoding=encoding)

    def read_all_cstrings(self, encoding: str, drop_empty: bool = False) -> list[str]:
        """
        Reads all null-terminated C strings from the buffer.
        """

        items: list[str] = []
        while self.data:
            items.append(self.read_cstring(encoding))

        if drop_empty:
            return [i for i in items if i]

        return items

    def read_remaining(self) -> bytes:
        """
        Reads out the remainder of this buffer.
        """

        ba = bytes(bytearray(self.data))
        self.data = deque()
        return ba



import pytest
from pg_purepy.conversion.hstore import get_hstore_converter

from tests.util import open_connection

pytestmark = pytest.mark.anyio


async def test_hstore():
    """
    Tests conversion for the hstore type.
    """

    async with open_connection() as conn:
        await conn.execute("create extension if not exists hstore;")
        converter = await get_hstore_converter(conn)
        assert converter, "wtf?"

        conn.add_converter(converter)

        result_1 = await conn.fetch_one("select 'a=>b'::hstore;")
        assert result_1
        assert result_1.data == [{"a": "b"}]

        result_2 = await conn.fetch_one("select $1::hstore;", {"c": "d"})
        assert result_2
        assert result_2.data == [{"c": "d"}]



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Self, override

if TYPE_CHECKING:
    from pg_purepy.connection import AsyncPostgresConnection


class PostgresqlError(Exception):
    """
    Base exception class all other exceptions are derived from.
    """


class ConnectionForciblyKilledError(PostgresqlError):
    """
    Raised when a connection is forcibly killed.
    """

    def __init__(self, conn: AsyncPostgresConnection):
        self._conn = conn

    @override
    def __str__(self) -> str:
        return f"Connection {self._conn!r} could not send the Terminate message"

    __repr__: Callable[[Self], str] = __str__


class ProtocolParseError(PostgresqlError):
    """
    Base exception class for all protocol parsing related errors.
    """


class MissingPasswordError(ProtocolParseError):
    """
    Raised when the server asks for a password, but we don't have one.
    """


class UnknownMessageError(ProtocolParseError):
    """
    Raised when an unknown message is returned.
    """


class IllegalStateError(ProtocolParseError):
    """
    Raised when an operation is attempted that would result in an illegal state.
    """


class ConnectionInTransactionWarning(ResourceWarning):
    """
    Raised when a connection is returned to a connection pool whilst it is still in a transaction.
    """


class MissingRowError(ValueError):
    """
    Raised when there's no row during a ``fetch_one`` operation.
    """



# THis function is copied from pg8000,
# at https://raw.githubusercontent.com/tlocke/pg8000/master/pg8000/dbapi.py
# It has been slightly modified to only allow named style.
#
# The following licence applies to this file:
#
# Copyright (c) 2007-2009, Mathieu Fenniak
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# mypy: ignore-errors
# pyright: basic

from itertools import count
from typing import Any, LiteralString


def convert_paramstyle(
    query: str, args: dict[str, Any]
) -> tuple[LiteralString, tuple[Any, ...]]:  # pragma: no cover
    # I don't see any way to avoid scanning the query string char by char,
    # so we might as well take that careful approach and create a
    # state-based scanner.  We'll use int variables for the state.
    OUTSIDE = 0  # outside quoted string
    INSIDE_SQ = 1  # inside single-quote string '...'
    INSIDE_QI = 2  # inside quoted identifier   "..."
    INSIDE_ES = 3  # inside escaped single-quote string, E'...'
    INSIDE_PN = 4  # inside parameter name eg. :name
    INSIDE_CO = 5  # inside inline comment eg. --

    in_quote_escape = False
    placeholders = []
    output_query = []
    ("$" + str(x) for x in count(1))
    state = OUTSIDE
    prev_c = None
    for i, c in enumerate(query):
        next_c = query[i + 1] if i + 1 < len(query) else None

        if state == OUTSIDE:
            if c == "'":
                output_query.append(c)
                state = INSIDE_ES if prev_c == "E" else INSIDE_SQ
            elif c == '"':
                output_query.append(c)
                state = INSIDE_QI
            elif c == "-":
                output_query.append(c)
                if prev_c == "-":
                    state = INSIDE_CO
            elif c == ":" and next_c not in ":=" and prev_c != ":":  # type: ignore
                # Same logic for : as in numeric parameters
                state = INSIDE_PN
                placeholders.append("")
            else:
                output_query.append(c)

        elif state == INSIDE_SQ:
            if c == "'":
                if in_quote_escape:
                    in_quote_escape = False
                else:
                    if next_c == "'":
                        in_quote_escape = True
                    else:
                        state = OUTSIDE
            output_query.append(c)

        elif state == INSIDE_QI:
            if c == '"':
                state = OUTSIDE
            output_query.append(c)

        elif state == INSIDE_ES:
            if c == "'" and prev_c != "\\":
                # check for escaped single-quote
                state = OUTSIDE
            output_query.append(c)

        elif state == INSIDE_PN:
            placeholders[-1] += c
            if next_c is None or (not next_c.isalnum() and next_c != "_"):
                state = OUTSIDE
                try:
                    pidx = placeholders.index(placeholders[-1], 0, -1)
                    output_query.append("$" + str(pidx + 1))
                    del placeholders[-1]
                except ValueError:
                    output_query.append("$" + str(len(placeholders)))

        elif state == INSIDE_CO:
            output_query.append(c)
            if c == "\n":
                state = OUTSIDE

        prev_c = c

    vals = tuple(args[p] for p in placeholders)
    return "".join(output_query), vals



import warnings
from contextlib import aclosing, suppress

import anyio
import pytest
import trio.testing
from anyio import BusyResourceError
from anyio.lowlevel import checkpoint
from pg_purepy.connection import open_database_connection
from pg_purepy.exc import MissingPasswordError, MissingRowError, ProtocolParseError
from pg_purepy.messages import (
    CommandComplete,
    DataRow,
    InvalidPasswordError,
    RecoverableDatabaseError,
    RowDescription,
)

from tests.util import (
    POSTGRES_ADDRESS,
    POSTGRES_PASSWORD,
    POSTGRES_PORT,
    POSTGRES_USERNAME,
    open_connection,
)

pytestmark = pytest.mark.anyio


async def test_basic_connection():
    """
    Tests opening a basic connection to the PostgreSQL server.
    """

    async with open_connection() as conn:
        assert conn.ready


@pytest.mark.skipif(not POSTGRES_PASSWORD, reason="no password provided")
async def test_connection_with_invalid_password():
    """
    Tests opening a connection with an invalid password.
    """

    with pytest.raises(InvalidPasswordError):
        async with open_database_connection(
            address_or_path=POSTGRES_ADDRESS,
            username=POSTGRES_USERNAME,
            port=POSTGRES_PORT,
            password="",
        ):
            pass


@pytest.mark.skipif(not POSTGRES_PASSWORD, reason="no password provided")
async def test_needs_password():
    """
    Tests opening a connection where a password is needed, but none is provided.
    """

    with pytest.raises(MissingPasswordError):
        async with open_database_connection(
            address_or_path=POSTGRES_ADDRESS,
            username=POSTGRES_USERNAME,
            port=POSTGRES_PORT,
        ):
            pass


async def test_basic_select():
    """
    Tests a basic SELECT statement.
    """

    async with open_connection() as conn:
        # fun fact: select null; returns text.
        result = [i async for i in conn.lowlevel_query("select null;")]
        assert len(result) == 3
        desc, row, count = result

        assert isinstance(desc, RowDescription)
        assert len(desc.columns) == 1
        assert desc.columns[0].name == "?column?"

        assert isinstance(row, DataRow)
        assert len(row.data) == 1
        assert row.data[0] is None

        assert isinstance(count, CommandComplete)
        assert count.row_count == 1


async def test_query_after_error():
    """
    Tests running a second query after a first query raises an error.
    """

    async with open_connection() as conn:
        # simple queries
        with pytest.raises(RecoverableDatabaseError) as e1:
            _ = await conn.fetch("select * from nonexistent;")

        assert e1.value.response.code == "42P01"

        row_1 = await conn.fetch_one("select 1;")
        assert len(row_1.data) == 1
        assert row_1.data[0] == 1

        with pytest.raises(RecoverableDatabaseError) as e2:
            await conn.fetch("select * from nonexistent where 'a' = :a;", a="a")

        assert e2.value.response.code == "42P01"

        row_2 = await conn.fetch_one("select 2 where 'a' = :a;", a="a")
        assert len(row_2.data) == 1
        assert row_2.data[0] == 2


async def test_query_with_positional_params():
    """
    Tests running a query with positional parameters.
    """

    async with open_connection() as conn:
        result = await conn.fetch_one("select $1::int4;", 7)

        assert result.data[0] == 7


async def test_query_with_params():
    """
    Tests running a query with parameters.
    """

    async with open_connection() as conn:
        row = await conn.fetch_one("select 1 where 'a' = :x;", x="a")
        assert row.data[0] == 1


async def test_get_row_count_no_rowcount():
    """
    Tests getting the row count of a command that doesn't return a row count.
    """

    async with open_connection() as conn:
        row_count = await conn.execute("checkpoint;")
        assert row_count == 0


async def test_multiple_queries_one_execute():
    """
    Tests executing multiple queries in one execute.
    """

    async with open_connection() as conn:
        async with aclosing(conn.lowlevel_query("select 1; select 2;")) as agen:
            results = [i async for i in agen]

        assert len(results) == 6
        row1, row2 = results[1], results[4]

        assert isinstance(row1, DataRow)
        assert row1.data[0] == 1
        assert isinstance(row2, DataRow)
        assert row2.data[0] == 2


async def test_empty_select_raises():
    """
    Tests that running ``fetch_one`` when the table has no rows raises an error.
    """

    async with open_connection() as conn:
        with pytest.raises(MissingRowError):  # noqa: PT012
            await conn.execute("create temporary table empty (_ int primary key);")
            await conn.fetch_one("select * from empty;")


## Transaction helper ##
async def test_transaction_status_manual():
    """
    Tests getting the transaction status during a query.
    """

    async with open_connection() as conn:
        await conn.execute("begin;")
        assert conn.in_transaction
        await conn.execute("rollback;")
        assert not conn.in_transaction


async def test_transaction_helper_normal():
    """
    Tests the transaction helper in a normal situation.
    """

    async with open_connection() as conn:
        async with conn.with_transaction():
            assert conn.in_transaction
            await conn.execute(
                "create temp table test_transaction_helper_normal (id int primary key);"
            )

        assert not conn.in_transaction
        result = await conn.fetch_one(
            "select count(*) from pg_tables where tablename = :name;",
            name="test_transaction_helper_normal",
        )

        assert result
        assert result.data[0] == 1


async def test_transaction_helper_error():
    """
    Tests the transaction helper in an error situation.
    """

    async with open_connection() as conn:
        with pytest.raises(ValueError):  # noqa: PT011, PT012
            async with conn.with_transaction():
                assert conn.in_transaction
                await conn.execute(
                    "create temp table test_transaction_helper_error (id int primary key);"
                )
                raise ValueError()

        assert not conn.in_transaction
        result = await conn.fetch_one(
            "select count(*) from pg_tables where tablename = :name;",
            name="test_transaction_helper_error",
        )

        assert result
        assert result.data[0] == 0


## Prepared statements ##
async def test_execute_prepared_statement_insert():
    """
    Tests executing a prepared statement with parameters.
    """

    async with open_connection() as conn:
        await conn.execute("""
            create temp table test_epsi (
                id int primary key generated always as identity, foo text not null
            );""")

        st_no_params = await conn.create_prepared_statement(
            name="test_epsp_1", query="insert into test_epsi(foo) values ('one');"
        )
        rows_no_params = await conn.execute(st_no_params)
        assert rows_no_params == 1

        st_with_params = await conn.create_prepared_statement(
            name="test_epsp_2", query="insert into test_epsi(foo) values ($1);"
        )
        rows_params = await conn.execute(st_with_params, "two")
        assert rows_params == 1

        count = await conn.fetch_one("select count(*) from test_epsi;")
        assert count
        assert count.data[0] == 2

        rows = await conn.fetch("select * from test_epsi;")
        assert len(rows) == 2
        assert rows[0].data[1] == "one"
        assert rows[1].data[1] == "two"


## Specific subcommands ##
async def test_insert():
    """
    Tests inserting data into a table.
    """

    async with open_connection() as conn:
        await conn.execute("""
            create temp table test_insert (
                id int primary key generated always as identity,
                foo text not null
            );
            """)
        row_count = await conn.execute("insert into test_insert(foo) values (:one);", one="test")
        assert row_count == 1
        result = await conn.fetch_one("select * from test_insert;")
        assert result
        assert result.data == [1, "test"]


async def test_unparameterised_insert():
    """
    Tests inserting static data.
    """

    async with open_connection() as conn:
        await conn.execute("""create temp table test_insert2 (
                id int primary key generated always as identity,
                foo text not null
            );""")
        row_count = await conn.execute("insert into test_insert2(foo) values ('test');")
        assert row_count == 1
        result = await conn.fetch_one("select * from test_insert2;")
        assert result
        assert result.data == [1, "test"]


async def test_update():
    """
    Tests updating data in a table.
    """

    async with open_connection() as conn:
        await conn.execute("""
            create temp table test_update (
                id int primary key generated always as identity, foo text not null
            );
            """)
        await conn.execute("insert into test_update(foo) values (:one);", one="test")
        pre_update = await conn.fetch_one("select * from test_update;")
        assert pre_update
        assert pre_update.data == [1, "test"]

        row_count = await conn.execute("update test_update set foo = :one;", one="test_2")
        assert row_count == 1
        post_update = await conn.fetch_one("select * from test_update;")
        assert post_update
        assert post_update.data == [1, "test_2"]


async def test_delete():
    """
    Tests deleting data in a table.
    """

    async with open_connection() as conn:
        await conn.execute("""
            create temp table test_delete (
                id int primary key generated always as identity,
                foo text not null
            );
            """)
        await conn.execute("insert into test_delete(foo) values (:one);", one="test")
        pre_delete = await conn.fetch("select * from test_delete;")
        assert len(pre_delete) == 1
        assert pre_delete[0].data == [1, "test"]

        row_count = await conn.execute("delete from test_delete;")
        assert row_count == 1

        post_delete = await conn.fetch("select * from test_delete;")
        assert len(post_delete) == 0


## Notices ##
async def test_notices():
    """
    Tests handling notices in the stream.
    """

    async with open_connection() as conn:
        with pytest.warns(UserWarning):  # noqa: PT030
            await conn.execute(
                "DO language plpgsql $$ BEGIN RAISE WARNING 'hello, world!'; END $$;"
            )

        with warnings.catch_warnings():
            await conn.execute("DO language plpgsql $$ BEGIN RAISE NOTICE 'hello, world!'; END $$;")


## Runtime Configuration ##
async def test_set_parameter():
    """
    Tests setting a runtime configuration parameter.
    """

    async with open_connection() as conn:
        await conn.execute("set application_name to 'test';")

        assert "application_name" in conn.connection_parameters
        assert conn.connection_parameters["application_name"] == "test"

        row = await conn.fetch_one("show application_name;")
        assert row.data[0] == conn.connection_parameters["application_name"]


async def test_set_illegal_parameter():
    """
    Tests setting an illegal parameter.
    """

    async with open_connection() as conn:
        with pytest.raises(ProtocolParseError):
            await conn.execute("set DateStyle to 'Postgres, MDY';")


## Misc ##
async def test_consume_all(anyio_backend: str):
    """
    Tests that getting the cached row count works.
    """

    async with open_connection() as conn, conn.query("select 1;") as query:
        rows = [r async for r in query]
        assert rows[0].data[0] == 1

        assert await query.consume_all() == 1

        if anyio_backend == "trio":
            with trio.testing.assert_checkpoints():
                assert await query.consume_all() == 1


async def test_seen_rows() -> None:
    """
    Tests the seen rows field.
    """

    async with (
        open_connection() as conn,
        conn.query("select * from ( values (1), (2), (3)) as d;") as query,
    ):
        assert query.seen_rows == 0
        it = aiter(query)
        _ = await anext(it)

        assert query.seen_rows == 1
        assert await query.consume_all() == 3
        assert query.seen_rows == 3


async def test_insert_into_not_null():
    """
    Tests inserting into a not-null table.
    """

    async with open_connection() as conn:
        await conn.execute("create temp table test (id int primary key);")

        with pytest.raises(RecoverableDatabaseError) as e:
            await conn.execute("insert into test(id) values (:n);", n=None)

        assert e.value.response.code == "23502"


async def test_max_rows():
    """
    Tests limiting the row count.
    """

    async with open_connection() as conn:
        await conn.execute("create temp table test (id int primary key);")
        await conn.execute("insert into test(id) values (1), (2), (3);")
        fetched = await conn.fetch("select * from test;", max_rows=1)
        assert len(fetched) == 1
        assert fetched[0].data[0] == 1


async def test_transaction_when_cancelled():
    """
    Tests handling transactions when cancelled.
    """

    async with open_connection() as conn:
        with anyio.CancelScope() as scope:
            async with conn.with_transaction():
                scope.cancel()
                await checkpoint()

        # synchronise
        await conn.execute("select 1;")

        assert not conn.in_transaction


async def test_returning_none_from_fetch_one():
    async with open_connection() as conn:
        res = await conn.fetch_one_or_none(
            "select 1 from pg_tables where schemaname = 'doesntexist';", return_none_on_empty=True
        )
        assert res is None


async def test_query_result_with_errors():
    # wow, 5 levels of indentation
    async with open_connection() as conn:
        with suppress(ValueError):
            async with conn.query("select * from ( values (1), (2), (3)) as d") as next:
                async for i in next:
                    assert i.data[0] == 1
                    raise ValueError

        assert (await conn.fetch_one("select 12345;")).data[0] == 12345


async def test_attempting_nested_queries():
    async with open_connection() as conn, conn.query("select * from ( values (3), (4)) as d;") as q:
        async for _ in q:
            with pytest.raises(BusyResourceError):
                await conn.fetch_one("select 2;")



from pg_purepy.connection import (
    AsyncPostgresConnection as AsyncPostgresConnection,
    QueryResult as QueryResult,
    RollbackTimeoutError as RollbackTimeoutError,
    open_database_connection as open_database_connection,
)
from pg_purepy.conversion import (
    Converter as Converter,
    EnumConverter as EnumConverter,
    SimpleFunctionConverter as SimpleFunctionConverter,
)
from pg_purepy.exc import (
    ConnectionForciblyKilledError as ConnectionForciblyKilledError,
    ConnectionInTransactionWarning as ConnectionInTransactionWarning,
    IllegalStateError as IllegalStateError,
    MissingPasswordError as MissingPasswordError,
    MissingRowError as MissingRowError,
    PostgresqlError as PostgresqlError,
    ProtocolParseError as ProtocolParseError,
    UnknownMessageError as UnknownMessageError,
)
from pg_purepy.messages import (
    ColumnDescription as ColumnDescription,
    ErrorOrNoticeResponse as ErrorOrNoticeResponse,
    ErrorResponseFieldType as ErrorResponseFieldType,
    RecoverableDatabaseError as RecoverableDatabaseError,
    RowDescription as RowDescription,
    UnrecoverableDatabaseError as UnrecoverableDatabaseError,
)
from pg_purepy.pool import (
    PooledDatabaseInterface as PooledDatabaseInterface,
    open_pool as open_pool,
)
from pg_purepy.protocol import SansIOClient as SansIOClient



"""
Converter type for array objects.
"""

from collections.abc import Iterable
from functools import partial
from typing import override

from pg_purepy.conversion.abc import ConversionContext, Converter
from pg_purepy.conversion.array_parse import _parse_array


class ArrayConverter[T](Converter[list[T]]):
    """
    Converts arrays to Python lists. This requires a subconverter which will be called to convert
    every value in the array.
    """

    def __init__(self, oid: int, subconverter: Converter[T], quote_inner: bool = False) -> None:
        """
        :param oid: The OID of the array type (not the base type!)
        :param subconverter: The converter for individual elements inside the array.
        :param quote_inner: When converting to PostgreSQL, if inner elements should be quoted.
        """
        self.oid = oid
        self._subconverter = subconverter
        self._quote_inner = quote_inner

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> list[T]:
        p = partial(self._subconverter.from_postgres, context)
        return _parse_array(data, p)

    @override
    def to_postgres(self, context: ConversionContext, data: Iterable[T]) -> str:
        converted = [
            self._subconverter.to_postgres(context, i) if i is not None else "NULL" for i in data
        ]
        if self._quote_inner:
            converted = ['"' + i + '"' for i in converted]

        return "{" + ",".join(converted) + "}"



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, override

from pg_purepy.conversion._parse_hstore import _parse_hstore, _serialize_hstore
from pg_purepy.conversion.abc import ConversionContext, Converter

if TYPE_CHECKING:
    from pg_purepy.pool import AsyncPostgresConnection


class HStoreConverter(Converter[Mapping[str, str]]):
    """
    Converter for the PostgreSQL hstore type.
    """

    def __init__(self, oid: int):
        self.oid = oid

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> Mapping[str, str]:
        return _parse_hstore(data)

    @override
    def to_postgres(self, context: ConversionContext, data: Mapping[str, str]) -> str:
        return _serialize_hstore(data)


async def get_hstore_converter(connection: AsyncPostgresConnection) -> HStoreConverter | None:
    """
    Gets the hstore converter for a connection. This may return None if the hstore extension
    is not currently enabled.
    """

    result = await connection.fetch_one(
        "select oid from pg_catalog.pg_type where pg_catalog.pg_type.typname = 'hstore';"
    )
    if not result:
        return None

    oid = result.data[0]
    assert oid

    return HStoreConverter(oid)



"""
The AnyIO implementation of the PostgreSQL client.
"""

from __future__ import annotations

import types
import warnings
from collections.abc import AsyncGenerator, AsyncIterator, Mapping
from contextlib import aclosing, asynccontextmanager
from os import PathLike, fspath
from ssl import SSLContext
from typing import (
    Any,
    Self,
    final,
    override,
)

import anyio
import anyio.lowlevel
import attrs
import structlog
from anyio import EndOfStream, ResourceGuard
from anyio.abc import ByteStream, SocketStream
from anyio.streams.tls import TLSStream

from pg_purepy.conversion.abc import Converter
from pg_purepy.dbapi import convert_paramstyle
from pg_purepy.exc import IllegalStateError, MissingRowError, PostgresqlError
from pg_purepy.messages import (
    BackendKeyData,
    BaseDatabaseError,
    BindComplete,
    CommandComplete,
    DataRow,
    ErrorOrNoticeResponse,
    PostgresMessage,
    PreparedStatementInfo,
    QueryResultMessage,
    wrap_error,
)
from pg_purepy.protocol import (
    NEED_DATA,
    SSL_MESSAGE,
    NeedData,
    ProtocolParseError,
    ReadyForQuery,
    SansIOClient,
    check_if_tls_accepted,
)

logger: structlog.stdlib.BoundLogger = structlog.get_logger(name=__name__)


class RollbackTimeoutError(PostgresqlError):
    """
    Raised when a ROLLBACK failed.
    """


class AsyncPostgresConnection:
    """
    An asynchronous connection to a PostgreSQL server. See method documentation for more info.

    This class should not be directly instantiated; instead, use :func:`.open_database_connection`.
    """

    def __init__(
        self,
        address_or_path: str | PathLike[str],
        port: int,
        ssl_context: SSLContext | None,
        stream: ByteStream,
        state: SansIOClient,
        block_transactions: bool = False,
    ):
        self._addr = address_or_path
        self._port = port
        self._ssl_context = ssl_context

        self._stream = stream
        self._protocol = state

        self._protocol_guard = ResourceGuard()

        # marks if the connection is dead, usually if a connection error happens during read/write.
        self._dead = False

        # backend PID as returned from postgresql.
        self._pid: int = -1
        self._secret_key: int = -1

        self._block_transactions = block_transactions

        self._logger = logger.bind(database=state.database)

    @property
    def ready(self) -> bool:
        """
        If this connection is ready to send another query.
        """

        return self._protocol.ready

    @property
    def in_transaction(self) -> bool:
        """
        If this connection is currently in a SQL transaction.
        """

        return self._protocol.in_transaction

    @property
    def dead(self) -> bool:
        """
        If this connection is unusable.
        """

        if self._dead:
            return True

        return self._protocol.dead

    @property
    def connection_parameters(self) -> Mapping[str, str]:
        """
        A read-only view of the current "connection parameters".

        This is a set of global variables set across a single PostgreSQL connection that control
        options such as the timezone or locale.
        """

        return types.MappingProxyType(self._protocol.connection_params)

    @property
    def server_timezone(self) -> str:
        """
        The raw timezone for this connection.
        """

        return self._protocol.timezone

    def add_converter(self, converter: Converter[Any]) -> None:
        """
        Registers a :class:`.Converter` with this connection.
        """

        self._protocol.add_converter(converter)

    @override
    def __repr__(self) -> str:
        return f"<{type(self).__name__} pid='{self._pid!r}'>"

    async def _read(self) -> bytes:
        try:
            return await self._stream.receive()
        except (ConnectionError, EndOfStream):
            self._dead = True
            raise

    async def _write(self, item: bytes) -> None:
        try:
            return await self._stream.send(item)
        except (ConnectionError, EndOfStream):
            self._dead = True
            raise

    async def _do_startup(self) -> None:
        data = self._protocol.do_startup()
        await self._write(data)

    async def _terminate(self) -> None:
        data = self._protocol.do_terminate()
        await self._write(data)
        self._dead = True

    async def _read_until_ready(
        self,
    ) -> AsyncGenerator[ErrorOrNoticeResponse | PostgresMessage | NeedData]:
        if self._protocol.ready:
            await anyio.lowlevel.checkpoint()
            return

        while True:
            while True:
                next_event = self._protocol.next_event()
                if next_event is NEED_DATA:
                    break

                if isinstance(next_event, ErrorOrNoticeResponse) and next_event.notice:
                    if next_event.severity == "WARNING":
                        err = wrap_error(next_event)
                        warnings.warn(str(err), stacklevel=2)

                elif isinstance(next_event, BackendKeyData):
                    self._logger = self._logger.bind(pid=next_event.pid)
                    logger.debug("BackendKeyData")
                    self._secret_key = next_event.secret_key
                    self._pid = next_event.pid

                yield next_event

                if isinstance(next_event, ReadyForQuery):
                    await anyio.lowlevel.checkpoint()
                    return

            to_send = self._protocol.get_needed_synchronisation()
            if to_send:
                await self._write(to_send)

            received = await self._read()
            self._protocol.receive_bytes(received)

    async def wait_until_ready(self) -> None:
        """
        Waits until the connection is ready.

        This will discard all other events on the connection.
        """

        async with aclosing(self._read_until_ready()) as gen:
            async for message in gen:
                if isinstance(message, ErrorOrNoticeResponse):
                    err = wrap_error(message)
                    raise err

    async def _wait_for_message[T](self, typ: type[T], *, wait_until_ready: bool = True) -> T:
        """
        Waits until a message of type ``typ`` arrives.

        This will wait until the ReadyForQuery message arrives to avoid requiring extra
        synchronisation if ``wait_until_ready`` is True. If it never arrives, this will deadlock!
        """

        message_found = None
        async with aclosing(self._read_until_ready()) as gen:
            async for item in gen:
                if isinstance(item, typ):
                    if not wait_until_ready:
                        return item

                    message_found = item

                elif isinstance(item, ErrorOrNoticeResponse):
                    raise wrap_error(item)

            if message_found is None:
                raise IllegalStateError(f"No message of type {typ} was yielded")

            return message_found

    ## Low-level API ##

    async def _do_create_prepared_statement(self, name: str, query: str) -> PreparedStatementInfo:
        to_send = self._protocol.do_create_prepared_statement(name=name, query_text=query)
        await self._write(to_send)
        return await self._wait_for_message(PreparedStatementInfo)

    async def create_prepared_statement(self, name: str, query: str) -> PreparedStatementInfo:
        """
        Creates a prepared statement. This is part of the low-level query API.

        :param name: The name of the prepared statement.
        :param query: The query to use.
        """

        with self._protocol_guard:
            return await self._do_create_prepared_statement(name=name, query=query)

    async def lowlevel_query(
        self,
        query: str | PreparedStatementInfo,
        *params: str,
        max_rows: int | None = None,
        **kwargs: dict[str, Any],
    ) -> AsyncGenerator[QueryResultMessage]:
        """
        Performs a query to the server.

        This is an *asynchronous generator*; it lazily fetches raw messages from the server without
        processing. You almost definitely want to use :meth:`.query` instead.
        """

        with self._protocol_guard:
            # always wait until ready! we do not like getting random messages from the last client
            # intermixed
            if not self._protocol.ready:
                await self.wait_until_ready()

            simple_query = all((
                not (params or kwargs),
                not isinstance(query, PreparedStatementInfo),
                max_rows is None,
            ))

            logger.debug("Executing query", query=query)
            if simple_query:
                data = self._protocol.do_simple_query(query)  # type: ignore
                await self._write(data)
            else:
                if not isinstance(query, PreparedStatementInfo):
                    real_query, new_params = convert_paramstyle(query, kwargs)
                    params = params + new_params
                    info = await self._do_create_prepared_statement(name="", query=real_query)
                else:
                    info = query

                bound_data = self._protocol.do_bind_execute(info, params, max_rows)
                await self._write(bound_data)
                # we need to get BindComplete because we need to yield the statement's
                # RowDescription out, for a more "consistent" view.
                await self._wait_for_message(BindComplete, wait_until_ready=False)
                # no error, so the query is gonna complete successfully

                if info.row_description:
                    yield info.row_description

            async with aclosing(self._read_until_ready()) as agen:
                async for message in agen:
                    if isinstance(message, ErrorOrNoticeResponse) and not message.notice:
                        err = wrap_error(message)
                        raise err

                    if isinstance(message, QueryResultMessage):
                        yield message

    ## Mid-level API. ##
    @asynccontextmanager
    async def query(
        self,
        query: str | PreparedStatementInfo,
        *params: str,
        max_rows: int | None = None,
        **kwargs: Any,
    ) -> AsyncGenerator[QueryResult]:
        """
        Mid-level query API.

        The ``query`` parameter can either be a string or a :class:`~.PreparedStatementInfo`, as
        returned from :meth:`.AsyncPostgresConnection.create_prepared_statement`. If it is a
        string, and it has parameters, they can be provided as either positional arguments or
        as keyword arguments. If it is a pre-prepared statement, and it has parameters,
        they must be provided as positional arguments.

        If keyword arguments are provided or a prepared statement is passed, an extended query with
        secure argument parsing will be used. Otherwise, a simple query will be used, which saves
        bandwidth over the extended query protocol.

        If the server is currently in a failed transaction, then your query will be ignored. Make
        sure to issue a rollback beforehand, if needed.

        This is an asynchronous context manager that yields a :class:`.QueryResult`, that can
        be asynchronously iterated over for the data rows of the query. Once all data rows have
        been iterated over, you can call :meth:`.QueryResult.consume_all` to get the total row
        count.

        If ``max_rows`` is specified, then the query will only return up to that many rows.
        Otherwise, an unlimited amount may potentially be returned.

        If this connection is currently executing another query, this method will raise a
        :class:`anyio.BusyResourceError`. The same is true of all other query methods that call
        this function; be sure to protect this type with a lock or use a connection pool for
        multiple simultaneous connections.
        """

        async with aclosing(
            self.lowlevel_query(query, *params, max_rows=max_rows, **kwargs),
        ) as agen:
            yield QueryResult(iterator=agen.__aiter__())
            # always wait
            await self.wait_until_ready()

    async def _safely_rollback(self, exc: BaseException | None) -> None:
        """
        Safely performs a rollback, even in the presence of an exception.
        """

        with anyio.move_on_after(5.0, shield=True) as scope:
            await self.execute("rollback;")

        if scope.cancel_called:
            with anyio.CancelScope(shield=True):
                await self._stream.aclose()
                self._dead = True

            raise RollbackTimeoutError(
                "Failed to rollback transaction in time, forcibly closing connection"
            ) from exc

    @asynccontextmanager
    async def with_transaction(self) -> AsyncGenerator[None]:
        """
        Asynchronous context manager that automatically opens and closes a transaction.
        """

        # TODO: this can be done at the type level?
        if self._block_transactions:
            raise ValueError(
                "This connection was already checked out from a pool in a "
                "transaction and this method should not be used."
            )

        try:
            await self.execute("begin;")
            yield
        except (Exception, anyio.get_cancelled_exc_class()) as e:
            await self._safely_rollback(e)
            raise
        else:
            await self.execute("commit;")

    ### DBAPI style methods ###
    async def fetch(
        self,
        query: str | PreparedStatementInfo,
        *params: Any,
        max_rows: int | None = None,
        **kwargs: Any,
    ) -> list[DataRow]:
        """
        Eagerly fetches the result of a query. This returns a list of :class:`~.DataRow` objects.

        If you wish to lazily load the results of a query, use :meth:`.query` instead.

        :param query: Either a :class:`str` that contains the query text,
                      or a :class:`~.PreparedStatementInfo` that represents a pre-prepared query.
        :param params: The positional arguments for the query.
        :param max_rows: The maximum rows to return.
        :param kwargs: The colon arguments for the query.
        """

        async with self.query(query, *params, max_rows=max_rows, **kwargs) as q:
            return [i async for i in q]

    async def fetch_one_or_none(
        self,
        query: str | PreparedStatementInfo,
        *params: Any,
        **kwargs: Any,
    ) -> DataRow | None:
        """
        Like :meth:`.fetch`, but only fetches one row.
        """

        row = await self.fetch(query, *params, **kwargs)

        try:
            return row[0]
        except IndexError:
            return None

    async def fetch_one(
        self, query: str | PreparedStatementInfo, *params: Any, **kwargs: Any
    ) -> DataRow:
        """
        Like :meth:`.fetch_one`, but raises :class:`.MissingRowError` if there's no row to fetch.
        """

        if res := await self.fetch_one_or_none(query, *params, **kwargs):
            return res

        raise MissingRowError()

    async def execute(
        self,
        query: str | PreparedStatementInfo,
        *params: Any,
        max_rows: int | None = None,
        **kwargs: Any,
    ) -> int:
        """
        Executes a query, returning its row count. This will discard all data rows.

        :param query: Either a :class:`str` that contains the query text,
                      or a :class:`~.PreparedStatementInfo` that represents a pre-prepared query.
        :param params: The positional arguments for the query.
        :param max_rows: The maximum rows to return.
        :param kwargs: The colon arguments for the query.
        """

        async with self.query(query, *params, max_rows=max_rows, **kwargs) as q:
            return await q.consume_all()


@attrs.define(kw_only=True)
@final
class QueryResult(AsyncIterator[DataRow]):
    """
    Wraps the extended result of a query.

    This is an *asynchronous iterator*; rows are retrieved with an ``async for``. The number of
    rows seen by the iterator so far can be found with the ``seen_rows`` field, or the total number
    of rows can be eagerly fetched with :meth:`.consume_all` (but this will discard any further
    data rows).
    """

    _iterator: AsyncIterator[PostgresMessage] = attrs.field(alias="iterator")
    _total_row_count: int = attrs.field(init=False, default=-1)

    #: The number of rows that this query has seen so far.
    seen_rows: int = attrs.field(init=False, default=0)

    @override
    def __aiter__(self) -> Self:
        return self

    @override
    async def __anext__(self) -> DataRow:
        # infinitely loops until we get a message we care about

        while True:
            next_message = await self._iterator.__anext__()
            if isinstance(next_message, DataRow):
                self.seen_rows += 1
                return next_message

            if isinstance(next_message, CommandComplete):
                # some messages don't have a row count, e.g. CREATE
                if next_message.row_count is None:
                    self._total_row_count = 0
                else:
                    self._total_row_count = next_message.row_count

                raise StopAsyncIteration

    async def consume_all(self) -> int:
        """
        Consumes all rows for this result and returns the total row count.
        """

        if self._total_row_count >= 0:
            await anyio.lowlevel.checkpoint()

        else:
            async for _ in self:
                pass

        return self._total_row_count


# TODO: support byte paths


async def _open_socket(
    address_or_path: str | PathLike[str],
    *,
    port: int = 5432,
    ssl_context: SSLContext | None = None,
) -> SocketStream | TLSStream:
    """
    Opens the socket to the PostgreSQL server.
    """

    address_or_path = fspath(address_or_path)
    sock: SocketStream | TLSStream
    if address_or_path.startswith("/"):  # pragma: no cover
        logger.debug("Opening connection", type="unix", path=address_or_path)
        sock = await anyio.connect_unix(address_or_path)
    else:
        logger.debug("Opening connection", type="tcp", address=address_or_path, port=port)
        sock = await anyio.connect_tcp(remote_host=address_or_path, remote_port=port)

    try:
        if ssl_context:
            logger.debug("Using TLS for connection")

            await sock.send(SSL_MESSAGE)
            response = await sock.receive(1)
            if not check_if_tls_accepted(response):
                raise ProtocolParseError("Requested TLS, but server said no")

            sock = await TLSStream.wrap(
                sock, hostname=address_or_path, ssl_context=ssl_context, standard_compatible=True
            )
    except BaseException:
        await sock.aclose()
        raise

    return sock


async def _open_connection(
    address_or_path: str | PathLike[str],
    username: str,
    *,
    port: int = 5432,
    password: str | None = None,
    database: str | None = None,
    ssl_context: SSLContext | None = None,
) -> tuple[ByteStream, AsyncPostgresConnection]:
    """
    Actual implementation of connection opening.
    """

    sock = await _open_socket(address_or_path, port=port, ssl_context=ssl_context)

    protocol = SansIOClient(username, database, password)
    conn = AsyncPostgresConnection(
        address_or_path,  # used for pool cancellation
        port,
        ssl_context,
        sock,
        protocol,
    )

    try:
        await conn._do_startup()
        await conn.wait_until_ready()
    except BaseException:
        await sock.aclose()
        raise

    return sock, conn


@asynccontextmanager
async def open_database_connection(
    address_or_path: str | PathLike[str],
    username: str,
    *,
    port: int = 5432,
    password: str | None = None,
    database: str | None = None,
    ssl_context: SSLContext | None = None,
) -> AsyncGenerator[AsyncPostgresConnection]:
    """
    Opens a new connection to the PostgreSQL database server. This is an asynchronous context
    manager.

    .. code-block:: python3

        async with open_database_connection("localhost", username="postgres") as db:
            ...

    Required parameters:

    :param address_or_path: The address of the server or the *absolute path* of its Unix socket.
    :param username: The username to authenticate with.

    Optional parameters:

    :param port: The port to connect to. Ignored for unix sockets.
    :param password: The password to authenticate with.
    :param database: The database to connect to. Defaults to the username.
    :param ssl_context: The SSL context to use for TLS connection. Enables TLS if specified.
    """

    sock, conn = await _open_connection(
        address_or_path=address_or_path,
        username=username,
        port=port,
        password=password,
        database=database,
        ssl_context=ssl_context,
    )

    async with sock:
        # this sucks but we send a Terminate in the normal case, a Terminate in the case of a
        # database error, and a regular socket/TLS close in all other cases.
        try:
            yield conn
            await conn._terminate()
        except BaseDatabaseError:
            await conn._terminate()
            raise



"""
Package containing the converterss for Python types to PostgreSQL types.
"""

from __future__ import annotations

import itertools
from collections.abc import Sequence
from functools import partial
from typing import TYPE_CHECKING

try:
    from ujson import dumps, loads as json_loads

    json_dumps = partial(dumps, ensure_ascii=False)
except ModuleNotFoundError:
    from json import dumps as json_dumps, loads as json_loads


from pg_purepy.conversion.abc import Converter as Converter
from pg_purepy.conversion.arrays import ArrayConverter as ArrayConverter
from pg_purepy.conversion.basics import (
    BoolConverter as BoolConverter,
    ByteaConverter as ByteaConverter,
    SimpleFunctionConverter as SimpleFunctionConverter,
    TextConverter as TextConverter,
)
from pg_purepy.conversion.dt import (
    STATIC_DATE_CONVERTER,
    STATIC_DATEA_CONVERTER,
    STATIC_TIME_CONVERTER,
    STATIC_TIMESTAMPNOTZ_CONVERTER,
    STATIC_TIMESTAMPNOTZA_CONVERTER,
    STATIC_TIMESTAMPTZ_CONVERTER,
    STATIC_TIMESTAMPTZA_CONVERTER,
)
from pg_purepy.conversion.enums import EnumConverter

if TYPE_CHECKING:
    from pg_purepy.protocol import SansIOClient


def _make_array_converters[T](
    oids: Sequence[int],
    cvs: Sequence[Converter[T]],
    *,
    quote_inner: bool = False,
) -> list[Converter[list[T]]]:
    return [
        ArrayConverter(oid, cv, quote_inner=quote_inner)
        for (oid, cv) in zip(oids, cvs, strict=True)
    ]


KNOWN_INT_OIDS = (20, 21, 23, 26, 27, 28, 29)
INT_CONVERTERS = [SimpleFunctionConverter(oid, int, str) for oid in KNOWN_INT_OIDS]
KNOWN_INTA_OIDS = (1016, 1005, 1007, 1028, 1010, 1011, 1012)
INTA_CONVERTERS = _make_array_converters(KNOWN_INTA_OIDS, INT_CONVERTERS)

KNOWN_TEXT_OIDS = (18, 19, 25)
STR_CONVERTERS = [TextConverter(oid) for oid in KNOWN_TEXT_OIDS]
KNOWN_STRA_OIDS = (1009, 1002, 1003)
STRA_CONVERTERS = _make_array_converters(KNOWN_STRA_OIDS, STR_CONVERTERS, quote_inner=True)

KNOWN_FLOAT_OIDS = (700, 701)
FLOAT_CONVERTERS = [SimpleFunctionConverter(oid, float, str) for oid in KNOWN_FLOAT_OIDS]
KNOWN_FLOATA_OIDS = (1021, 1022)
FLOATA_CONVERTERS = _make_array_converters(KNOWN_FLOATA_OIDS, FLOAT_CONVERTERS)

KNOWN_JSON_OIDS = (114, 3802)
JSON_CONVERTERS = [SimpleFunctionConverter(oid, json_loads, json_dumps) for oid in KNOWN_JSON_OIDS]
KNOWN_JSONA_OIDS = (199, 3807)
JSONA_CONVERTERS = _make_array_converters(KNOWN_JSONA_OIDS, JSON_CONVERTERS)

STATIC_BOOLEAN_CONVERTER = BoolConverter()
STATIC_BOOLEANA_CONVERTER = ArrayConverter(1000, STATIC_BOOLEAN_CONVERTER)
STATIC_BYTES_CONVERTER = ByteaConverter()


def apply_default_converters(protocol: SansIOClient) -> None:
    """
    Applies the default converter objects to a protocol.
    """

    # fmt: off
    for cv in itertools.chain(
        INT_CONVERTERS, INTA_CONVERTERS,
        STR_CONVERTERS, STRA_CONVERTERS,
        FLOAT_CONVERTERS, FLOATA_CONVERTERS,
        JSON_CONVERTERS, JSONA_CONVERTERS,
    ):
        protocol.add_converter(cv)
    # fmt: on

    protocol.add_converter(STATIC_BOOLEAN_CONVERTER)
    protocol.add_converter(STATIC_BOOLEANA_CONVERTER)

    protocol.add_converter(STATIC_BYTES_CONVERTER)

    # datetime converters
    protocol.add_converter(STATIC_TIMESTAMPTZ_CONVERTER)
    protocol.add_converter(STATIC_TIMESTAMPTZA_CONVERTER)

    protocol.add_converter(STATIC_TIMESTAMPNOTZ_CONVERTER)
    protocol.add_converter(STATIC_TIMESTAMPNOTZA_CONVERTER)

    protocol.add_converter(STATIC_DATE_CONVERTER)
    protocol.add_converter(STATIC_DATEA_CONVERTER)

    protocol.add_converter(STATIC_TIME_CONVERTER)
    protocol.add_converter(STATIC_DATEA_CONVERTER)


__all__ = ("ArrayConverter", "Converter", "EnumConverter", "apply_default_converters")



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Literal, override

import dateutil.parser
import whenever

from pg_purepy.conversion.abc import Converter
from pg_purepy.conversion.arrays import ArrayConverter

if TYPE_CHECKING:
    from pg_purepy.protocol import ConversionContext


type PostgresInfinity = Literal["infinity", "-infinity"]
type PostgresTimestampTz = whenever.OffsetDateTime | PostgresInfinity
type PostgresTimestampWithoutTz = whenever.PlainDateTime | PostgresInfinity


class TimestampTzConverter(Converter[PostgresTimestampTz]):
    """
    Converts from a TIMESTAMP WITH TIMEZONE to an Arrow object.
    """

    oid = 1184

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> PostgresTimestampTz:
        if data == "infinity" or data == "-infinity":
            return data

        # TIMESTAMPTZ are stored in UTC, and are converted to the server's timezone on retrieval.
        # So we provide the returned date in the server's timezone.
        parsed = dateutil.parser.isoparse(data)

        # can't directly pass the datetime as it'll complain about UTC.
        return whenever.OffsetDateTime.parse_iso(parsed.isoformat())

    @override
    def to_postgres(
        self,
        context: ConversionContext,
        data: PostgresTimestampTz,
    ) -> str:
        match data:
            case "infinity":
                return "infinity"

            case "-infinity":
                return "-infinity"

            case whenever.OffsetDateTime():
                return data.format_iso()


STATIC_TIMESTAMPTZ_CONVERTER = TimestampTzConverter()
STATIC_TIMESTAMPTZA_CONVERTER = ArrayConverter(118, STATIC_TIMESTAMPTZ_CONVERTER)


class TimestampNoTzConverter(Converter[PostgresTimestampWithoutTz]):
    """
    Converts from a TIMESTAMP WITHOUT TIMEZONE to an Arrow object.
    """

    oid = 1114

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> PostgresTimestampWithoutTz:
        if data == "infinity" or data == "-infinity":
            return data

        return whenever.PlainDateTime.parse_iso(data)

    @override
    def to_postgres(self, context: ConversionContext, data: PostgresTimestampWithoutTz) -> str:
        # we can once again just do isoformat, because postgres will parse it, go "nice offset.
        # fuck you" and completely discard it.
        #
        # postgres@127:postgres> select '2021-07-13T22:16:36+01:00'::timestamp;
        # +---------------------+
        # | timestamp           |
        # |---------------------|
        # | 2021-07-13 22:16:36 |
        # +---------------------+

        match data:
            case "infinity":
                return data

            case "-infinity":
                return data

            case whenever.PlainDateTime():
                return data.format_iso()


STATIC_TIMESTAMPNOTZ_CONVERTER = TimestampNoTzConverter()
STATIC_TIMESTAMPNOTZA_CONVERTER = ArrayConverter(1115, STATIC_TIMESTAMPNOTZ_CONVERTER)


class DateConverter(Converter[datetime.date]):
    """
    Converts from a DATE to a :class:`datetime.date` object.
    """

    oid = 1082

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> datetime.date:
        return datetime.date.fromisoformat(data)

    @override
    def to_postgres(self, context: ConversionContext, data: datetime.date) -> str:
        return data.isoformat()


STATIC_DATE_CONVERTER = DateConverter()
STATIC_DATEA_CONVERTER = ArrayConverter(1182, STATIC_DATE_CONVERTER)


class TimeConverter(Converter[datetime.time]):
    """
    Converts from a TIME WITHOUT TIMEZONE to a :class`datetime.time` object.
    """

    oid = 1083

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> datetime.time:
        return datetime.time.fromisoformat(data)

    @override
    def to_postgres(self, context: ConversionContext, data: datetime.time) -> str:
        return data.isoformat()


STATIC_TIME_CONVERTER = TimeConverter()
STATIC_TIMEA_CONVERTER = ArrayConverter(1183, STATIC_TIME_CONVERTER)



import pytest

from tests.util import open_connection

pytestmark = pytest.mark.anyio


async def test_data_row_to_dict():
    async with open_connection() as conn:
        row = await conn.fetch_one('select 1 as "column";')
        assert row

        row_dict = row.to_dict()
        assert "column" in row_dict
        assert row_dict["column"] == 1



import os
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager

from pg_purepy import AsyncPostgresConnection, open_database_connection

POSTGRES_ADDRESS = os.environ.get("POSTGRES_ADDRESS", "127.0.0.1")
POSTGRES_PORT: int = int(os.environ.get("POSTGRES_PORT", "5432"))
POSTGRES_USERNAME = os.environ.get("POSTGRES_USERNAME", "postgres")
POSTGRES_PASSWORD = os.environ.get("POSTGRES_PASSWORD", "")  # lol!
POSTGRES_DATABASE = os.environ.get("POSTGRES_DATABASE", "postgres")


@asynccontextmanager
async def open_connection() -> AsyncGenerator[AsyncPostgresConnection]:
    async with open_database_connection(
        address_or_path=POSTGRES_ADDRESS,
        port=POSTGRES_PORT,
        username=POSTGRES_USERNAME,
        password=POSTGRES_PASSWORD,
        database=POSTGRES_DATABASE,
    ) as conn:
        yield conn



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, override

from pg_purepy.conversion.abc import Converter

if TYPE_CHECKING:
    from pg_purepy.protocol import ConversionContext


class SimpleFunctionConverter[ConvType](Converter[ConvType]):
    """
    A simple converter that calls fn_from_pg to convert to a Python type, and fn_to_pg to convert
    from a Python type.
    """

    def __init__(
        self, oid: int, fn_from_pg: Callable[[str], ConvType], fn_to_pg: Callable[[ConvType], str]
    ) -> None:
        self.oid = oid

        self._from_pg = fn_from_pg
        self._to_pg = fn_to_pg

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> ConvType:
        return self._from_pg(data)

    @override
    def to_postgres(self, context: ConversionContext, data: ConvType) -> str:
        return self._to_pg(data)


class BoolConverter(Converter[bool]):
    """
    Converter that converts for booleans.
    """

    oid = 16

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> bool:
        return data == "t"

    @override
    def to_postgres(self, context: ConversionContext, data: bool) -> str:
        return "true" if data else "false"


class TextConverter(Converter[str]):
    """
    Text converter for default text types.
    """

    def __init__(self, oid: int):
        self.oid = oid

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> str:
        return data

    @override
    def to_postgres(self, context: ConversionContext, data: str) -> str:
        return str(data)


class ByteaConverter(Converter[bytes]):
    """
    Converter that turns bytes objects into bytea objects.
    """

    oid = 17

    @override
    def from_postgres(self, context: ConversionContext, data: str) -> Any:
        # bytea data is hex escaped.
        prefix, rest = data[0:2], data[2:]
        assert prefix == r"\x"

        return bytes.fromhex(rest)

    @override
    def to_postgres(self, context: ConversionContext, data: bytes) -> str:
        return r"\x" + data.hex()



from enum import Enum
from typing import cast

import pytest
from pg_purepy import EnumConverter, UnrecoverableDatabaseError

from tests.hilevel import open_pool

pytestmark = pytest.mark.anyio


class ExampleEnum(Enum):
    ONE = "one"
    TWO = "two"
    THREE = "three"


async def test_get_type_oid():
    """
    Tests getting a type OID.
    """
    async with open_pool(conn_count=1) as pool:
        oid = await pool.find_oid_for_type("int4")
        assert oid == 23


async def test_adding_converter():
    """
    Tests adding a converter to connections.
    """
    async with open_pool(conn_count=1) as pool:
        await pool.execute("drop type if exists test_ace_t;")
        await pool.execute("create type test_ace_t as enum ('one', 'two', 'three');")
        oid = await pool.find_oid_for_type("test_ace_t")
        assert oid, "wtf?"

        converter = EnumConverter(oid, ExampleEnum)

        pool.add_converter(converter)

        row_1 = await pool.fetch_one("select 'one'::test_ace_t;")
        assert row_1
        assert row_1[0] == ExampleEnum.ONE

        # kill connection, to test that the converter is added after reconnect
        with pytest.raises(UnrecoverableDatabaseError) as e:
            await pool.execute("select pg_terminate_backend(pg_backend_pid());")

        assert e.value.response.code == "57P01"
        row_2 = await pool.fetch_one("select 'two'::test_ace_t;")
        assert row_2
        assert row_2[0] == ExampleEnum.TWO


async def test_adding_converter_and_array():
    """
    Tests adding a converter and its array converter at the same time.
    """
    async with open_pool(conn_count=1) as pool:
        await pool.execute("drop type if exists test_acaa;")
        await pool.execute("create type test_acaa as enum ('one', 'two', 'three');")

        oid_row = await pool.fetch_one("select oid from pg_type where typname = 'test_acaa';")
        assert oid_row
        oid: int = cast(int, oid_row.data[0])
        converter = EnumConverter(oid, ExampleEnum)
        pool.add_converter(converter)

        result_1 = await pool.fetch_one("select '{\"one\"}'::test_acaa[];")
        assert result_1
        assert result_1[0] == "{one}"

        # safe, this uses a set and will just overwrite the old converter cleanly.
        await pool.add_converter_with_array(converter)
        result_2 = await pool.fetch_one("select '{\"one\"}'::test_acaa[];")
        assert result_2
        assert result_2[0] == [ExampleEnum.ONE]



from __future__ import annotations

import abc

import attrs


class Converter[IntoType](metaclass=abc.ABCMeta):
    """
    Base class for all conversion classes. Implement this to create a custom converter.
    """

    #: The OID of the PostgreSQL type this converter uses.
    oid: int

    @abc.abstractmethod
    def from_postgres(self, context: ConversionContext, data: str) -> IntoType:
        """
        Converts ``data`` from the PostgreSQL string representation to a Python type.

        :param context: The conversion context this converter was invoked in.
        :param data: The raw string data.
        :return: Any Python object that resulted from the conversion.
        """

    @abc.abstractmethod
    def to_postgres(self, context: ConversionContext, data: IntoType) -> str:
        """
        Converts ``data`` from the Python type to the PostgreSQL string representation.

        :param context: The conversion context this converter was invoked in.
        :param data: The Python object that needs to be converted.
        :return: The string data that will be used in a query string.
        """


@attrs.define(slots=True, frozen=False)
class ConversionContext:
    """
    Information that may be needed during conversion from PostgreSQL types.
    """

    #: The encoding of the client.
    client_encoding: str = attrs.field()

    #: The raw timezone of the server.
    timezone: str = attrs.field(default="UTC")



import pytest

from tests.util import open_connection

pytestmark = pytest.mark.anyio


async def test_converters_int():
    """
    Tests int converters.
    """
    async with open_connection() as conn:
        for cast in (
            "int2",
            "int4",
            "int8",
            "oid",
        ):
            result_out = await conn.fetch(f"select 1::{cast};")
            assert result_out[0].data[0] == 1

            await conn.execute(f"create temp table test_int_{cast} (id {cast} primary key);")
            await conn.execute(f"insert into test_int_{cast} values ($1);", 39)
            result_in = await conn.fetch_one(f"select id from test_int_{cast};")
            assert result_in

            assert result_in.data[0] == 39


async def test_converter_bool():
    """
    Tests the boolean converter.
    """
    async with open_connection() as conn:
        true = await conn.fetch_one("select true;")
        assert true
        assert true.data[0] is True

        false = await conn.fetch_one("select false;")
        assert false
        assert false.data[0] is False

        await conn.execute("create temp table test_bool (id int primary key, value boolean);")
        await conn.execute(
            "insert into test_bool(id, value) values (1, :t), (2, :f);", t=True, f=False
        )
        rows = await conn.fetch("select value from test_bool;")
        assert rows[0].data[0] is True
        assert rows[1].data[0] is False


async def test_converter_bytea():
    """
    Tests the bytea converter.
    """
    async with open_connection() as conn:
        # trick: both in and out conversion
        ba = await conn.fetch_one(r"select $1::bytea;", b"\x00\x01")
        assert ba
        assert isinstance(ba.data[0], bytes)
        assert ba.data[0] == b"\x00\x01"



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager

from pg_purepy.pool import PooledDatabaseInterface, open_pool as _open_pool

from tests.util import (
    POSTGRES_ADDRESS,
    POSTGRES_DATABASE,
    POSTGRES_PASSWORD,
    POSTGRES_PORT,
    POSTGRES_USERNAME,
)


@asynccontextmanager
async def open_pool(conn_count: int | None = None) -> AsyncGenerator[PooledDatabaseInterface]:
    async with _open_pool(
        connection_count=conn_count,
        address_or_path=POSTGRES_ADDRESS,
        port=POSTGRES_PORT,
        username=POSTGRES_USERNAME,
        password=POSTGRES_PASSWORD,
        database=POSTGRES_DATABASE,
    ) as c:
        yield c



# This code is copied from pg8000, at
# https://github.com/tlocke/pg8000/blob/e4a8b49825baa475dd65547e966767a9a98d6954/pg8000/converters.py#L454-L502
#
# The following licence applies to this file:
#
# Copyright (c) 2007-2009, Mathieu Fenniak
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# mypy: ignore-errors
# pyright: basic

from collections.abc import Callable
from enum import Enum


class ArrayState(Enum):
    InString = 1
    InEscape = 2
    InValue = 3
    Out = 4


def _parse_array[T](data: str, adapter: Callable[[str], T]) -> list[T]:
    state = ArrayState.Out
    stack = [[]]
    val = []
    for c in data:
        if state == ArrayState.InValue:
            if c in ("}", ","):
                value = "".join(val)
                stack[-1].append(None if value == "NULL" else adapter(value))
                state = ArrayState.Out
            else:
                val.append(c)

        if state == ArrayState.Out:
            if c == "{":
                a = []
                stack[-1].append(a)
                stack.append(a)
            elif c == "}":
                stack.pop()
            elif c == ",":
                pass
            elif c == '"':
                val = []
                state = ArrayState.InString
            else:
                val = [c]
                state = ArrayState.InValue

        elif state == ArrayState.InString:
            if c == '"':
                stack[-1].append(adapter("".join(val)))
                state = ArrayState.Out
            elif c == "\\":
                state = ArrayState.InEscape
            else:
                val.append(c)
        elif state == ArrayState.InEscape:
            val.append(c)
            state = ArrayState.InString

    return stack[0][0]



import pytest

from tests.util import open_connection

pytestmark = pytest.mark.anyio


async def test_basic_arrays():
    """
    Tests basic array functionality.
    """
    async with open_connection() as conn:
        await conn.execute("create temp table test_ba_t (arr int4[] primary key);")
        await conn.execute("insert into test_ba_t(arr) values (:arr)", arr=[1, 2, 3])

        data_row = await conn.fetch_one("select * from test_ba_t;")
        assert data_row
        assert data_row[0] == [1, 2, 3]


async def test_string_array():
    """
    Tests string arrays with commas.
    """
    async with open_connection() as conn:
        await conn.execute("create temp table test_sa_t (arr text[] primary key);")
        await conn.execute("insert into test_sa_t values (:arr)", arr=["one, two", "three,"])
        data_row = await conn.fetch_one("select * from test_sa_t;")

        assert data_row
        assert data_row[0] == ["one, two", "three,"]



from itertools import product

import pytest

from tests.util import open_connection

pytestmark = pytest.mark.anyio


JSON_TYPES = ["json", "jsonb"]
JSON_EXAMPLES = [('{"abc": "def"}', {"abc": "def"}), ("[1, 2, 3]", [1, 2, 3])]


async def test_loading_json():
    async with open_connection() as conn:
        for type_, (text, expected) in product(JSON_TYPES, JSON_EXAMPLES):
            result = await conn.fetch_one(f"select '{text}'::{type_}")
            assert result
            assert result.data[0] == expected


async def test_inserting_json():
    async with open_connection() as conn:
        await conn.execute("create temp table test (j jsonb primary key);")
        await conn.execute("insert into test (j) values ($1)", {"key": "value"})
        res = await conn.fetch_one("select j from test;")
        assert res
        assert res.data[0] == {"key": "value"}



"""
Tests date/time related converters.
"""

import datetime
from typing import cast

import pytest
from whenever import OffsetDateTime

from tests.util import open_connection

pytestmark = pytest.mark.anyio


async def test_timestamptz_converter():
    """
    Tests converting to/from a TIMESTAMP WITH TIMEZONE.
    """
    async with open_connection() as conn:
        await conn.execute("set timezone to 'UTC';")
        assert conn.server_timezone == "UTC"

        await conn.execute("create temp table test_ttz_in (id int primary key, dt timestamptz);")
        juno = OffsetDateTime.parse_iso("2011-08-05T16:25:00Z")
        # assert juno.zon == conn.server_timezone

        await conn.execute(
            "insert into test_ttz_in(id, dt) values (1, :dt);",
            dt=OffsetDateTime.parse_iso("2011-08-05T16:25:00Z"),
        )

        row_1 = await conn.fetch_one("select dt from test_ttz_in;")
        assert row_1
        assert isinstance(row_1.data[0], OffsetDateTime)
        assert row_1.data[0] == juno

        await conn.execute("set timezone to 'Europe/Helsinki';")
        row_2 = await conn.fetch_one("select dt from test_ttz_in;")
        assert row_2

        dt = cast(OffsetDateTime, row_2.data[0])
        assert dt.hour == 19
        assert dt.hour != juno.hour
        assert dt == juno


async def test_date_converter():
    """
    Tests converting to/from a DATE.
    """
    async with open_connection() as conn:
        await conn.execute("create temp table test_dci (dt date primary key);")
        lolk = datetime.date(2015, 8, 11)
        await conn.execute("insert into test_dci(dt) values (:date)", date=lolk)
        result = await conn.fetch_one("select * from test_dci;")
        assert result

        assert result.data[0] == lolk


async def test_time_converter():
    """
    Tests converting to/from a TIME.
    """
    async with open_connection() as conn:
        await conn.execute("create temp table test_tc (dt time primary key);")
        rn = datetime.time(1, 35, 33)
        await conn.execute("insert into test_tc values (:time);", time=rn)
        result = await conn.fetch_one("select * from test_tc;")

        assert result
        assert result.data[0] == rn


async def test_dt_infinity():
    """
    Tests using infinity in datetimes.
    """
    async with open_connection() as conn:
        await conn.execute("create temp table test_dti_in (dt timestamptz);")
        await conn.execute("insert into test_dti_in(dt) values (:inf);", inf="infinity")

        result = await conn.fetch_one("select * from test_dti_in;")
        assert result
        assert result.data[0] == "infinity"



from typing import Any

import anyio
import pytest
from pg_purepy import ConnectionInTransactionWarning, UnrecoverableDatabaseError

from tests.hilevel import open_pool

pytestmark = pytest.mark.anyio


async def test_successful_pool_usage():
    """
    Tests a successful pool usage.
    """

    results: set[Any] = set()

    async with open_pool(conn_count=3) as p:

        async def execute(num: int) -> None:
            # the sleep ensures that the connections actually run in parallel, as trio will
            # schedule the tasks before that sleep expires.
            result = await p.fetch_one(f"select pg_sleep(0.25), {num};")
            assert result

            results.add(result.data[1])

        async with anyio.create_task_group() as nursery:
            for i in range(3):
                nursery.start_soon(execute, i + 1)

        assert results == {1, 2, 3}


async def test_automatic_reconnection():
    """
    Tests the pool automatically reconnecting dead connections.
    """

    async with open_pool(conn_count=1) as p:
        with pytest.raises(UnrecoverableDatabaseError) as e:
            await p.execute("select pg_terminate_backend(pg_backend_pid());")

        assert e.value.response.code == "57P01"

        assert (await p.execute("select 2;")) == 1


async def test_pool_transactions():
    """
    Tests connection pool transactions.
    """

    async with open_pool(conn_count=1) as p:
        with pytest.raises(RuntimeError):  # noqa: PT012
            async with p.checkout_in_transaction() as conn:
                assert conn.in_transaction
                await conn.execute(
                    "create table test_transaction_helper_error (id int primary key);"
                )
                raise RuntimeError()

        result = await p.fetch_one(
            "select count(*) from pg_tables where tablename = :name;",
            name="test_transaction_helper_error",
        )

        assert result.data[0] == 0


async def test_transaction_rollback():
    """
    Tests that transactions are automatically rolled back.
    """

    async with open_pool(conn_count=1) as p:
        with pytest.warns(ConnectionInTransactionWarning):
            await p.execute("begin;")


async def test_cancellation():
    """
    Tests query cancellation integration.
    """

    async with open_pool(conn_count=1) as p:
        # timeout block causes an automatic cancel after a second, which (should) send the
        # cancellation request.
        with anyio.move_on_after(1.0):
            await p.execute("select pg_sleep_for('1 minutes');")

        # longer cooldown but still less than a minutee to let the cancellation happen
        with anyio.move_on_after(5.0) as s:
            await p.execute("select 1;")

        assert not s.cancel_called


async def test_cancellation_with_transactions():
    """
    Tests cancellation alongside transactions.
    """

    async with open_pool(conn_count=1) as p:
        with anyio.move_on_after(1.0):
            async with p.checkout_in_transaction() as conn:
                await conn.execute("select pg_sleep_for('1 minutes');")

        # eww
        assert not next(iter(p._raw_connections)).in_transaction



from pg_purepy.protocol import (
    NEED_DATA,
    AuthenticationCompleted,
    ParameterStatus,
    SansIOClient,
)


def test_tricky_packet_truncation():
    """
    Tests when packets are truncated.
    """
    state = SansIOClient(username="postgres")
    state.do_startup()

    # This has one missing NUL from the end.
    state.receive_bytes(b"\x52\x00\x00\x00\x08\x00\x00\x00")
    assert state.next_event() == NEED_DATA

    # The null byte is prepended here.
    state.receive_bytes(
        b"\x00\x53\x00\x00\x00\x1f\x61\x70\x70\x6c\x69\x63\x61\x74\x69\x6f\x6e\x5f"
        b"\x6e"
        b"\x61\x6d\x65\x00\x70\x67\x2d\x70\x75\x72\x65\x70\x79\x00"
    )

    assert isinstance(state.next_event(), AuthenticationCompleted)
    status = state.next_event()
    assert isinstance(status, ParameterStatus)
    assert status.name == "application_name"
    assert status.value == "pg-purepy"
